# Copyright 2018 Rubrik, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

"""
This module contains the Rubrik SDK Data_Management class.
"""
import re
from .api import Api
from .exceptions import CDMVersionException, InvalidParameterException, InvalidTypeException


_API = Api


class Data_Management(_API):
    """This class contains methods related to backup and restore operations for the various objects managed by the Rubrik cluster."""

    def on_demand_snapshot(self, object_name, object_type, sla_name='current', fileset=None, host_os=None, sql_host=None, sql_instance=None, sql_db=None, timeout=15):  # pylint: ignore
        """Initiate an on-demand snapshot.

        Arguments:
            object_name {str} -- The name of the Rubrik object to take a on-demand snapshot of.
            object_type {str} -- The Rubrik object type you want to backup. (choices: {vmware, physical_host, ahv, mssql})

        Keyword Arguments:
            sla_name {str} -- The SLA Domain name you want to assign the on-demand snapshot to. By default, the currently assigned SLA Domain will be used. (default: {'current'})
            fileset {str} -- The name of the Fileset you wish to backup. Only required when taking a on-demand snapshot of a physical host. (default: {'None'})
            host_os {str} -- The operating system for the physical host. Only required when taking a on-demand snapshot of a physical host. (default: {'None'}) (choices: {Linux, Windows})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            tuple -- When object_type is vmware, the full API response for `POST /v1/vmware/vm/{ID}/snapshot` and the job status URL which can be used to monitor progress of the snapshot. (api_response, job_status_url)
            tuple -- When object_type is physical_host, the full API response for `POST /v1/fileset/{}/snapshot` and the job status URL which can be used to monitor progress of the snapshot. (api_response, job_status_url)

        """

        valid_object_type = ['vmware', 'physical_host', 'ahv', 'mssql_db']
        valid_host_os_type = ['Linux', 'Windows']

        if object_type not in valid_object_type:
            raise InvalidParameterException("The on_demand_snapshot() `object_type` argument must be one of the following: {}.".format(
                valid_object_type))

        if host_os is not None:
            if host_os not in valid_host_os_type:
                raise InvalidParameterException("The on_demand_snapshot() `host_os` argument must be one of the following: {}.".format(
                    valid_host_os_type))

        if object_type == 'vmware':
            self.log("on_demand_snapshot: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type, timeout=timeout)

            if sla_name == 'current':
                self.log(
                    "on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain assigned to the vSphere VM '{}'.".format(object_name))

                vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)
                sla_id = vm_summary['effectiveSlaDomainId']

            elif sla_name != 'current':
                self.log("on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
                sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

            config = {}
            config['slaId'] = sla_id

            self.log("on_demand_snapshot: Initiating snapshot for the vSphere VM '{}'.".format(object_name))
            api_request = self.post('v1', '/vmware/vm/{}/snapshot'.format(vm_id), config, timeout)

            snapshot_status_url = api_request['links'][0]['href']

        elif object_type == 'ahv':

            self.log("on_demand_snapshot: Searching the Rubrik cluster for the AHV VM '{}'.".format(object_name))

            vm_id = self.object_id(object_name, object_type, timeout=timeout)

            if sla_name == 'current':
                self.log(
                    "on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain assigned to the AHV VM '{}'.".format(object_name))

                vm_summary = self.get('internal', '/nutanix/vm/{}'.format(vm_id), timeout)
                sla_id = vm_summary['effectiveSlaDomainId']

            elif sla_name != 'current':
                self.log("on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
                sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

            config = {}
            config['slaId'] = sla_id

            self.log("on_demand_snapshot: Initiating snapshot for the AHV VM '{}'.".format(object_name))
            api_request = self.post('internal', '/nutanix/vm/{}/snapshot'.format(vm_id), config, timeout)

            snapshot_status_url = api_request['links'][0]['href']

        elif object_type == 'mssql_db':

            self.log(
                "on_demand_snapshot: Searching the Rubrik cluster for the MS SQL '{}'.".format(object_name))

            mssql_host = self.object_id(sql_host, 'physical_host', timeout=timeout)

            mssql_instance = self.get(
                'v1', '/mssql/instance?primary_cluster_id=local&root_id={}'.format(mssql_host), timeout)

            for instance in mssql_instance['data']:
                if instance['name'] == sql_instance:
                    sql_db_id = instance['id']

            mssql_db = self.get('v1', '/mssql/db?primary_cluster_id=local&instance_id={}'.format(sql_db_id), timeout)

            for db in mssql_db['data']:
                if db['name'] == sql_db:
                    mssql_id = db['id']

            if sla_name == 'current':
                self.log(
                    "on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain assigned to the MS SQL '{}'.".format(object_name))

                mssql_summary = self.get('v1', '/mssql/db/{}'.format(mssql_id), timeout)
                sla_id = mssql_summary['effectiveSlaDomainId']

            elif sla_name != 'current':
                self.log("on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
                sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

            config = {}
            config['slaId'] = sla_id

            self.log("on_demand_snapshot: Initiating snapshot for the MS SQL '{}'.".format(object_name))
            api_request = self.post('v1', '/mssql/db/{}/snapshot'.format(mssql_id), config, timeout)

            snapshot_status_url = api_request['links'][0]['href']

        elif object_type == 'physical_host':
            if host_os is None:
                raise InvalidParameterException(
                    "The on_demand_snapshot() `host_os` argument must be populated when taking a Physical host snapshot.")
            elif fileset is None:
                raise InvalidParameterException(
                    "The on_demand_snapshot() `fileset` argument must be populated when taking a Physical host snapshot.")

            self.log("on_demand_snapshot: Searching the Rubrik cluster for the Physical Host '{}'.".format(object_name))
            host_id = self.object_id(object_name, object_type, timeout=timeout)

            self.log("on_demand_snapshot: Searching the Rubrik cluster for the Fileset Template '{}'.".format(fileset))
            fileset_template_id = self.object_id(fileset, 'fileset_template', host_os, timeout=timeout)

            self.log("on_demand_snapshot: Searching the Rubrik cluster for the full Fileset.")
            api_endpoint = '/fileset?primary_cluster_id=local&host_id={}&is_relic=false&template_id={}'.format(
                host_id, fileset_template_id)
            fileset_summary = self.get('v1', api_endpoint, timeout=timeout)

            if fileset_summary['total'] == 0:
                raise InvalidParameterException(
                    "The Physical Host '{}' is not assigned to the '{}' Fileset.".format(
                        object_name, fileset))

            fileset_id = fileset_summary['data'][0]['id']

            if sla_name == 'current':
                sla_id = fileset_summary['data'][0]['effectiveSlaDomainId']
            elif sla_name != 'current':
                self.log("on_demand_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
                sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

            config = {}
            config['slaId'] = sla_id

            self.log("on_demand_snapshot: Initiating snapshot for the Physical Host '{}'.".format(object_name))
            api_request = self.post('v1', '/fileset/{}/snapshot'.format(fileset_id), config, timeout)

            snapshot_status_url = api_request['links'][0]['href']

        return (api_request, snapshot_status_url)

    def object_id(self, object_name, object_type, host_os=None, timeout=15):
        """Get the ID of a Rubrik object by providing its name.

        Arguments:
            object_name {str} -- The name of the Rubrik object whose ID you wish to lookup.
            object_type {str} -- The object type you wish to look up. (choices: {vmware, sla, vmware_host, physical_host, fileset_template, managed_volume, aws_native, vcenter})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            str -- The ID of the provided Rubrik object.
        """

        valid_object_type = [
            'vmware',
            'sla',
            'vmware_host',
            'physical_host',
            'fileset_template',
            'managed_volume',
            'mssql_db',
            'mssql_instance',
            'vcenter',
            'ahv',
            'aws_native']

        if object_type not in valid_object_type:
            raise InvalidParameterException("The object_id() object_type argument must be one of the following: {}.".format(
                valid_object_type))

        if object_type == 'fileset_template':
            if host_os is None:
                raise InvalidParameterException("You must provide the Fileset Tempalte OS type.")
            elif host_os not in ['Linux', 'Windows']:
                raise InvalidParameterException("The host_os must be either 'Linux' or 'Windows'.")

        if object_type == 'sla':
            if object_name.upper() == "FOREVER" or object_name.upper() == "UNPROTECTED":
                return "UNPROTECTED"

        api_call = {
            "vmware": {
                "api_version": "v1",
                "api_endpoint": "/vmware/vm?primary_cluster_id=local&is_relic=false&name={}".format(object_name)
            },
            "sla": {
                "api_version": "v1",
                "api_endpoint": "/sla_domain?primary_cluster_id=local&name={}".format(object_name)
            },
            "vmware_host": {
                "api_version": "v1",
                "api_endpoint": "/vmware/host?primary_cluster_id=local"
            },
            "fileset_template": {
                "api_version": "v1",
                "api_endpoint": "/fileset_template?primary_cluster_id=local&operating_system_type={}&name={}".format(host_os, object_name)
            },
            "managed_volume": {
                "api_version": "internal",
                "api_endpoint": "/managed_volume?is_relic=false&primary_cluster_id=local&name={}".format(object_name)
            },
            "ahv": {
                "api_version": "internal",
                "api_endpoint": "/nutanix/vm?primary_cluster_id=local&is_relic=false&name={}".format(object_name)
            },
            "mssql_db": {
                "api_version": "v1",
                "api_endpoint": "/mssql/db?primary_cluster_id=local&is_relic=false&instance_id={}".format(object_name)
            },
            "mssql_instance": {
                "api_version": "v1",
                "api_endpoint": "/mssql/instance?primary_cluster_id=local&root_id={}".format(object_name)
            },
            "aws_native": {
                "api_version": "internal",
                "api_endpoint": "/aws/account?name={}".format(object_name)
            },
            "vcenter": {
                "api_version": "v1",
                "api_endpoint": "/vmware/vcenter"
            }
        }

        if object_type == 'physical_host':
            if self.minimum_installed_cdm_version(5.0, timeout) is True:
                filter_field_name = "name"
            else:
                filter_field_name = "hostname"

            api_call["physical_host"] = {
                "api_version": "v1",
                "api_endpoint": "/host?primary_cluster_id=local&{}={}".format(filter_field_name, object_name)
            }

        self.log("object_id: Getting the object id for the {} object '{}'.".format(object_type, object_name))
        api_request = self.get(
            api_call[object_type]["api_version"],
            api_call[object_type]["api_endpoint"],
            timeout=timeout)

        if api_request['total'] == 0:
            raise InvalidParameterException("The {} object '{}' was not found on the Rubrik cluster.".format(
                object_type, object_name))
        elif api_request['total'] > 0:
            object_ids = []
            # Define the "object name" to search for
            if object_type == 'physical_host':
                name_value = filter_field_name
            else:
                name_value = 'name'

            for item in api_request['data']:
                if item[name_value] == object_name:
                    object_ids.append(item['id'])

            if len(object_ids) > 1:
                raise InvalidParameterException(
                    "Multiple {} objects named '{}' were found on the Rubrik cluster. Unable to return a specific object id.".format(object_type, object_name))
            elif len(object_ids) == 0:
                raise InvalidParameterException(
                    "The {} object '{}' was not found on the Rubrik cluster.".format(object_type, object_name))
            else:
                return object_ids[0]

            raise InvalidParameterException(
                "The {} object '{}' was not found on the Rubrik cluster.".format(object_type, object_name))

    def assign_sla(self, object_name, sla_name, object_type, log_backup_frequency_in_seconds=None, log_retention_hours=None, copy_only=None, timeout=30):  # pytest: ignore
        """Assign a Rubrik object to an SLA Domain.

        Arguments:
            object_name {str} -- The name of the Rubrik object you wish to assign to an SLA Domain.
            sla_name {str} -- The name of the SLA Domain you wish to assign an object to. To exclude the object from all SLA assignments use `do not protect` as the `sla_name`. To assign the selected object to the SLA of the next higher level object use `clear` as the `sla_name`.
            object_type {str} -- The Rubrik object type you want to assign to the SLA Domain. (choices: {vmware, mssql_host})

        Keyword Arguments:
            log_backup_frequency_in_seconds {str} -- The MSSQL Log Backup frequency you'd like to specify with the SLA. Required when the `object_type` is `mssql_host`. (default {None})
            log_retention_hours {int} -- The MSSQL Log Retention frequency you'd like to specify with the SLA. Required when the `object_type` is `mssql_host`. (default {None})
            copy_only {int} -- Take Copy Only Backups with MSSQL. Required when the `object_type` is `mssql_host`. (default {None})
            timeout {bool} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            str -- No change required. The vSphere VM '`object_name`' is already assigned to the '`sla_name`' SLA Domain.
            dict -- The full API reponse for `POST /internal/sla_domain/{sla_id}/assign`.
        """

        valid_object_type = ['vmware', 'mssql_host']

        if object_type not in valid_object_type:
            raise InvalidParameterException(
                "The assign_sla() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type == "mssql_host":
            if log_backup_frequency_in_seconds is None or log_retention_hours is None or copy_only is None:
                raise InvalidParameterException(
                    "When the object_type is 'mssql_host' the 'log_backup_frequency_in_seconds', 'log_retention_hours', 'copy_only' paramaters must be populated.")
        # Determine if 'do not protect' or 'clear' are the SLA Domain Name
        do_not_protect_regex = re.findall('\\bdo not protect\\b', sla_name, flags=re.IGNORECASE)
        clear_regex = re.findall('\\bclear\\b', sla_name, flags=re.IGNORECASE)

        if len(do_not_protect_regex) > 0:
            sla_id = "UNPROTECTED"
        elif len(clear_regex) > 0:
            sla_id = 'INHERIT'
        else:
            self.log("assign_sla: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
            sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

        if object_type == 'vmware':
            self.log("assign_sla: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type, timeout=timeout)

            self.log("assign_sla: Determing the SLA Domain currently assigned to the vSphere VM '{}'.".format(object_name))
            vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)

            if sla_id == vm_summary['configuredSlaDomainId']:
                return "No change required. The vSphere VM '{}' is already assigned to the '{}' SLA Domain.".format(
                    object_name, sla_name)
            else:
                self.log("assign_sla: Assigning the vSphere VM '{}' to the '{}' SLA Domain.".format(object_name, sla_name))

                config = {}
                config['managedIds'] = [vm_id]

                return self.post("internal", "/sla_domain/{}/assign".format(sla_id), config, timeout)

        elif object_type == 'mssql_host':

            host_id = ''
            mssql_id = ''
            db_sla_lst = []

            self.log('Searching the Rubrik cluster for the current hosts.')
            current_hosts = self.get(
                'v1',
                '/host?operating_system_type=Windows&primary_cluster_id=local',
                timeout=timeout)

            # After 5.0, "hostname" is a deprecated field in the results that are returned in "current_hosts"
            if self.minimum_installed_cdm_version(5.0):
                current_hosts_name = "name"
            else:
                current_hosts_name = "hostname"

            for rubrik_host in current_hosts['data']:
                if rubrik_host[current_hosts_name] == object_name:
                    host_id = rubrik_host['id']

            if(host_id):
                self.log("assign_sla: Searching the Rubrik cluster for the MSSQL Instance '{}'.".format(object_name))
                mssql_instances = self.get('v1', '/mssql/instance?root_id={}'.format(host_id), timeout=timeout)

                for mssql_instance in mssql_instances['data']:
                    mssql_id = mssql_instance['id']
                    mssql_instance_name = mssql_instance['name']

                    self.log(
                        "assign_sla: Determing the SLA Domain currently assigned to the MSSQL Instance '{}'.".format(mssql_instance_name))

                    mssql_summary = self.get('v1', '/mssql/instance/{}'.format(mssql_id), timeout=timeout)

                    if (sla_id == mssql_summary['configuredSlaDomainId'] and log_backup_frequency_in_seconds == mssql_summary['logBackupFrequencyInSeconds'] and
                            log_retention_hours == mssql_summary['logRetentionHours'] and copy_only == mssql_summary['copyOnly']):
                        return "No change required. The MSSQL Instance '{}' is already assigned to the '{}' SLA Domain with the following log settings:" \
                               " log_backup_frequency_in_seconds: {}, log_retention_hours: {} and copy_only: {}.".format(
                                   object_name, sla_name, log_backup_frequency_in_seconds, log_retention_hours, copy_only)

                    else:
                        self.log(
                            "assign_sla: Assigning the MSSQL Instance '{}' to the '{}' SLA Domain.".format(
                                object_name, sla_name))

                        config = {}
                        if log_backup_frequency_in_seconds is not None:
                            config['logBackupFrequencyInSeconds'] = log_backup_frequency_in_seconds
                        if log_retention_hours is not None:
                            config['logRetentionHours'] = log_retention_hours
                        if copy_only is not None:
                            config['copyOnly'] = copy_only

                        config['configuredSlaDomainId'] = sla_id

                        patch_resp = self.patch("v1", "/mssql/instance/{}".format(mssql_id), config, timeout)
                        db_sla_lst.append(patch_resp)
            else:
                raise InvalidParameterException(
                    "Host ID not found for instance '{}'").format(object_name)

            return db_sla_lst

    def vsphere_live_mount(self, vm_name, date='latest', time='latest', host='current', remove_network_devices=False, power_on=True, timeout=15):  # pylint: ignore
        """Live Mount a vSphere VM from a specified snapshot. If a specific date and time is not provided, the last snapshot taken will be used.

        Arguments:
            vm_name {str} -- The name of the vSphere VM to Live Mount.

        Keyword Arguments:
            date {str} -- The date of the snapshot you wish to Live Mount formated as `Month-Day-Year` (ex: 1-15-2014). If `latest` is specified, the last snapshot taken will be used. (default: {'latest'})
            time {str} -- The time of the snapshot you wish to Live Mount formated formated as `Hour:Minute AM/PM` (ex: 1:30 AM). If `latest` is specified, the last snapshot taken will be used. (default: {'latest'})
            host {str} -- The hostname or IP address of the ESXi host to Live Mount the VM on. By default, the current host will be used. (default: {'current'})
            remove_network_devices {bool} -- Flag that determines whether to remove the network interfaces from the Live Mounted VM. Set to `True` to remove all network interfaces. (default: {False})
            power_on {bool} -- Flag that determines whether the VM should be powered on after the Live Mount. Set to `True` to power on the VM. Set to `False` to mount the VM but not power it on. (default: {True})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            dict -- The full response of `POST /v1/vmware/vm/snapshot/{snapshot_id}/mount`.
        """

        if isinstance(remove_network_devices, bool) is False:
            raise InvalidTypeException("The 'remove_network_devices' argument must be True or False.")
        elif isinstance(power_on, bool) is False:
            raise InvalidTypeException("The 'power_on' argument must be True or False.")
        elif date != 'latest' and time == 'latest' or date == 'latest' and time != 'latest':
            raise InvalidParameterException(
                "The date and time arguments most both be 'latest' or a specific date and time.")

        self.log("vsphere_live_mount: Searching the Rubrik cluster for the vSphere VM '{}'.".format(vm_name))
        vm_id = self.object_id(vm_name, 'vmware', timeout=timeout)

        self.log("vsphere_live_mount: Getting a list of all Snapshots for vSphere VM '{}'.".format(vm_name))
        vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)

        if date == 'latest' and time == 'latest':
            number_of_snapshots = len(vm_summary['snapshots'])
            snapshot_id = vm_summary['snapshots'][number_of_snapshots - 1]['id']
        else:
            self.log("vsphere_live_mount: Converting the provided date/time into UTC.")
            snapshot_date_time = self._date_time_conversion(date, time)

            current_snapshots = {}

            for snapshot in vm_summary['snapshots']:
                current_snapshots[snapshot['id']] = snapshot['date']

            self.log("vsphere_live_mount: Searching for the provided snapshot.")
            for id, date_time in current_snapshots.items():
                if snapshot_date_time in date_time:
                    snapshot_id = id

        try:
            snapshot_id
        except NameError:
            raise InvalidParameterException("The vSphere VM '{}' does not have a snapshot taken on {} at {}.".format(
                vm_name, date, time))
        else:
            if host == 'current':
                host_id = vm_summary['hostId']
            else:
                host_id = self.object_id(host, 'vmware_host', timeout=timeout)

            config = {}
            config['hostId'] = host_id
            config['removeNetworkDevices'] = remove_network_devices
            config['powerOn'] = power_on

            self.log(
                "vsphere_live_mount: Live Mounting the snapshot taken on {} at {} for vSphere VM '{}'.".format(
                    date,
                    time,
                    vm_name))

            return self.post('v1', '/vmware/vm/snapshot/{}/mount'.format(snapshot_id), config, timeout)

    def vsphere_instant_recovery(self, vm_name, date='latest', time='latest', host='current', remove_network_devices=False, power_on=True, disable_network=False, keep_mac_addresses=False, preserve_moid=False, timeout=15):  # pylint: ignore
        """Instantly recover a vSphere VM from a provided snapshot. If a specific date and time is not provided, the last snapshot taken will be used.

        Arguments:
            vm_name {str} -- The name of the VM to Instantly Recover.

        Keyword Arguments:
            date {str} -- The date of the snapshot you wish to Instantly Recover formated as `Month-Day-Year` (ex: 1-15-2014). If 'latest' is specified, the last snapshot taken will used. (default: {'latest'})
            time {str} -- The time of the snapshot you wish to Instantly Recover formated formated as `Hour:Minute AM/PM`  (ex: 1:30 AM). If 'latest' is specified, the last snapshot taken will be used. (default: {'latest'})
            host {str} -- The hostname or IP address of the ESXi host to Instantly Recover the VM on. By default, the current host will be used. (default: {'current'})
            remove_network_devices {bool} -- Flag that determines whether to remove the network interfaces from the Instantly Recovered VM. Set to `True` to remove all network interfaces. (default: {False})
            power_on {bool} -- Flag that determines whether the VM should be powered on after Instant Recovery. Set to `True` to power on the VM. Set to `False` to instantly recover the VM but not power it on. (default: {True})
            disable_network {bool} -- Sets the state of the network interfaces when the VM is instantly recovered. Use `False` to enable the network interfaces. Use `True` to disable the network interfaces. Disabling the interfaces can prevent IP conflicts. (default: {False})
            keep_mac_addresses {bool} -- Flag that determines whether the MAC addresses of the network interfaces on the source VM are assigned to the new VM. Set to `True` to assign the original MAC addresses to the new VM. Set to `False` to assign new MAC addresses. When 'remove_network_devices' is set to `True`, this property is ignored. (default: {False})
            preserve_moid {bool} -- Flag that determines whether to preserve the MOID of the source VM in a restore operation. Use `True` to keep the MOID of the source. Use `False` to assign a new moid. (default: {False})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})

        Returns:
            dict -- The full response of `POST /v1/vmware/vm/snapshot/{snapshot_id}/instant_recover`.
        """

        if isinstance(remove_network_devices, bool) is False:
            raise InvalidTypeException("The 'remove_network_devices' argument must be True or False.")
        elif isinstance(power_on, bool) is False:
            raise InvalidTypeException("The 'power_on' argument must be True or False.")
        elif isinstance(disable_network, bool) is False:
            raise InvalidTypeException("The 'disable_network' argument must be True or False.")
        elif isinstance(keep_mac_addresses, bool) is False:
            raise InvalidTypeException("The 'keep_mac_addresses' argument must be True or False.")
        elif isinstance(preserve_moid, bool) is False:
            raise InvalidTypeException("The 'preserve_moid' argument must be True or False.")
        elif date != 'latest' and time == 'latest' or date == 'latest' and time != 'latest':
            raise InvalidParameterException(
                "The date and time arguments most both be 'latest' or a specific date and time.")

        self.log("vsphere_instant_recovery: Searching the Rubrik cluster for the vSphere VM '{}'.".format(vm_name))
        vm_id = self.object_id(vm_name, 'vmware', timeout=timeout)

        self.log("vsphere_instant_recovery: Getting a list of all Snapshots for vSphere VM '{}'.".format(vm_name))
        vm_summary = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)

        if date == 'latest' and time == 'latest':
            number_of_snapshots = len(vm_summary['snapshots'])
            snapshot_id = vm_summary['snapshots'][number_of_snapshots - 1]['id']
        else:
            self.log("vsphere_instant_recovery: Converting the provided date/time into UTC.")
            snapshot_date_time = self._date_time_conversion(date, time)

            current_snapshots = {}
            for snapshot in vm_summary['snapshots']:
                current_snapshots[snapshot['id']] = snapshot['date']

            self.log("vsphere_instant_recovery: Searching for the provided snapshot.")
            for id, date_time in current_snapshots.items():
                if snapshot_date_time in date_time:
                    snapshot_id = id

        try:
            snapshot_id
        except NameError:
            raise InvalidParameterException(
                "The vSphere VM '{}' does not have a snapshot taken on {} at {}.".format(
                    vm_name, date, time))
        else:
            if host == 'current':
                host_id = vm_summary['hostId']
            else:
                host_id = self.object_id(host, 'vmware_host', timeout=timeout)

            config = {}
            config['hostId'] = host_id
            config['removeNetworkDevices'] = remove_network_devices
            config['powerOn'] = power_on
            config['disableNetwork'] = disable_network
            config['keepMacAddresses'] = keep_mac_addresses
            config['preserveMoid'] = preserve_moid

            self.log("vsphere_instant_recovery: Instantly Recovering the snapshot taken on {} at {} for vSphere VM '{}'.".format(
                date,
                time,
                vm_name))

            return self.post('v1', '/vmware/vm/snapshot/{}/instant_recover'.format(snapshot_id), config, timeout)

    def _date_time_conversion(self, date, time, timeout=30):
        """All date values returned by the Rubrik API are stored in Coordinated Universal Time (UTC)
        and need to be converted to the timezone configured on the Rubrik cluster in order to match
        the values provided by the end user in various functions. This internal function will handle that
        conversion process.

        Arguments:
            date {str} -- A date value formated as `Month-Day-Year` (ex: 1/15/2014).
            time {str} -- A time value formated as `Hour:Minute AM/PM` (ex: 1:30 AM).

        Returns:
            str -- A combined date/time value formated as `Year-Month-DayTHour:Minute` where Hour:Minute is on the 24-hour clock (ex : 2014-1-15T01:30).
        """

        from datetime import datetime
        import pytz

        # Validate the Date formating
        try:
            snapshot_date = datetime.strptime(date, '%m-%d-%Y')
        except ValueError:
            raise InvalidParameterException(
                "The date argument '{}' must be formatd as 'Month-Date-Year' (ex: 8-9-2018).".format(date))
        # Validate the Time formating
        try:
            snapshot_time = datetime.strptime(time, '%I:%M %p')
        except ValueError:
            raise InvalidParameterException(
                "The time argument '{}' must be formatd as 'Hour:Minute AM/PM' (ex: 2:57 AM).".format(time))

        self.log("_date_time_conversion: Getting the Rubrik cluster timezone.")
        cluster_summary = self.get('v1', '/cluster/me', timeout=timeout)
        cluster_timezone = cluster_summary['timezone']['timezone']

        self.log("_date_time_conversion: Converting the provided time to the 24-hour clock.")
        snapshot_time_24_hour_clock = datetime.strftime(snapshot_time, "%H:%M")

        self.log("_date_time_conversion: Creating a combined date/time variable.")
        snapshot_datetime = datetime.strptime('{} {}'.format(
            date, snapshot_time_24_hour_clock), '%m-%d-%Y %H:%M')

        # Add Timezone to snapshot_datetime Variable
        timezone = pytz.timezone(cluster_timezone)
        snapshot_datetime = timezone.localize(snapshot_datetime)

        self.log("_date_time_conversion: Converting the time to UTC.\n")
        utc_timezone = pytz.UTC
        snapshot_datetime = snapshot_datetime.astimezone(utc_timezone)

        return snapshot_datetime.strftime('%Y-%m-%dT%H:%M')

    def pause_snapshots(self, object_name, object_type, timeout=180):
        """Pause all snapshot activity for the provided object.

        Arguments:
            object_name {str} -- The name of the Rubrik object to pause snapshots for.
            object_type {str} -- The Rubrik object type you wish to pause snaphots on. (choices: {vmware})

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {180})

        Returns:
            str -- No change required. The '`object_type`' '`object_name`' is already paused.
            dict -- The full API response for `PATCH /v1/vmware/vm/{vm_id}`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            raise InvalidParameterException("The pause_snapshots() object_type argument must be one of the following: {}.".format(
                valid_object_type))

        if object_type == 'vmware':

            self.log("pause_snapshots: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type, timeout=timeout)

            self.log("pause_snapshots: Determing the current pause state of the vSphere VM '{}'.".format(object_name))
            api_request = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)

            if api_request['blackoutWindowStatus']['isSnappableBlackoutActive']:
                return "No change required. The {} VM '{}' is already paused.".format(object_type, object_name)
            else:
                self.log("pause_snapshots: Pausing Snaphots for the vSphere VM '{}'.".format(object_name))
                config = {}
                config['isVmPaused'] = True

                return self.patch('v1', '/vmware/vm/{}'.format(vm_id), config, timeout)

    def resume_snapshots(self, object_name, object_type, timeout=180):
        """Resume all snapshot activity for the provided object.

        Arguments:
            object_name {str} -- The name of the Rubrik object to resume snapshots for.
            object_type {str} -- The Rubrik object type you wish to resume snaphots on. (choices: {vmware})

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {180})

        Returns:
            str -- No change required. The 'object_type' object 'object_name' is currently not paused.
            dict -- The full response for `PATCH /v1/vmware/vm/{vm_id}`.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            raise InvalidParameterException("The resume_snapshots() object_type argument must be one of the following: {}.".format(
                valid_object_type))

        if object_type == 'vmware':

            self.log("resume_snapshots: Searching the Rubrik cluster for the vSphere VM '{}'.".format(object_name))
            vm_id = self.object_id(object_name, object_type, timeout=timeout)

            self.log("resume_snapshots: Determing the current pause state of the vSphere VM '{}'.".format(object_name))
            api_request = self.get('v1', '/vmware/vm/{}'.format(vm_id), timeout=timeout)

            if not api_request['blackoutWindowStatus']['isSnappableBlackoutActive']:
                return "No change required. The '{}' object '{}' is currently not paused.".format(
                    object_type, object_name)
            else:
                self.log("resume_snapshots: Resuming Snaphots for the vSphere VM '{}'.".format(object_name))
                config = {}
                config['isVmPaused'] = False

                return self.patch('v1', '/vmware/vm/{}'.format(vm_id), config, timeout)

    def begin_managed_volume_snapshot(self, name, timeout=30):
        """Open a managed volume for writes. All writes to the managed volume until the snapshot is ended will be part of its snapshot.

        Arguments:
            name {str} -- The name of the Managed Volume to begin the snapshot on.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {30})

        Returns:
            str -- No change required. The Managed Volume '`name`' is already assigned in a writeable state.
            dict -- The full API response for `POST /managed_volume/{id}/begin_snapshot`.
        """

        self.log("begin_managed_volume_snapshot: Searching the Rubrik cluster for the Managed Volume '{}'.".format(name))
        managed_volume_id = self.object_id(name, 'managed_volume', timeout=timeout)

        self.log("begin_managed_volume_snapshot: Determing the state of the Managed Volume '{}'.".format(name))
        managed_volume_summary = self.get('internal', '/managed_volume/{}'.format(managed_volume_id), timeout=timeout)

        if not managed_volume_summary['isWritable']:
            self.log("begin_managed_volume_snapshot: Setting the Managed Volume '{}' to a writeable state.".format(name))
            return self.post('internal', '/managed_volume/{}/begin_snapshot'.format(managed_volume_id),
                             config={}, timeout=timeout)
        else:
            return "No change required. The Managed Volume '{}' is already assigned in a writeable state.".format(name)

    def end_managed_volume_snapshot(self, name, sla_name='current', timeout=30):
        """Close a managed volume for writes. A snapshot will be created containing all writes since the last begin snapshot call.

        Arguments:
            name {str} -- The name of the Managed Volume to end snapshots on.

        Keyword Arguments:
            sla_name {str} -- The SLA Domain name you want to assign the snapshot to. By default, the currently assigned SLA Domain will be used. (default: {'current'})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {30})

        Returns:
            str -- No change required. The Managed Volume `name` is already assigned in a read only state.
            dict -- The full API response for `POST /managed_volume/{id}/end_snapshot`.
        """

        self.log("end_managed_volume_snapshot: Searching the Rubrik cluster for the Managed Volume '{}'.".format(name))
        managed_volume_id = self.object_id(name, 'managed_volume', timeout=timeout)

        self.log("end_managed_volume_snapshot: Determing the state of the Managed Volume '{}'.".format(name))
        managed_volume_summary = self.get("internal", "/managed_volume/{}".format(managed_volume_id), timeout=timeout)

        if not managed_volume_summary['isWritable']:
            return "No change required. The Managed Volume 'name' is already assigned in a read only state."

        if sla_name == 'current':
            self.log(
                "end_managed_volume_snapshot: Searching the Rubrik cluster for the SLA Domain assigned to the Managed Volume '{}'.".format(name))
            if managed_volume_summary['slaAssignment'] == 'Unassigned' or managed_volume_summary['effectiveSlaDomainId'] == 'UNPROTECTED':
                raise InvalidParameterException(
                    "The Managed Volume '{}' does not have a SLA assigned currently assigned. You must populate the sla_name argument.".format(name))
            config = {}
        else:
            self.log("end_managed_volume_snapshot: Searching the Rubrik cluster for the SLA Domain '{}'.".format(sla_name))
            sla_id = self.object_id(sla_name, 'sla', timeout=timeout)

            config = {}
            config['retentionConfig'] = {}
            config['retentionConfig']['slaId'] = sla_id

        return self.post("internal", "/managed_volume/{}/end_snapshot".format(managed_volume_id), config, timeout)

    def get_sla_objects(self, sla, object_type, timeout=15):
        """Retrieve the name and ID of a specific object type.

        Arguments:
            sla {str} -- The name of the SLA Domain you wish to search.
            object_type {str} -- The object type you wish to search the SLA for. (choices: {vmware})

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster. (default: {15})

        Returns:
            dict -- The `name:id` of each object in the provided SLA Domain.
        """

        valid_object_type = ['vmware']

        if object_type not in valid_object_type:
            raise InvalidParameterException(
                "The get_sla_object() object_type argument must be one of the following: {}.".format(valid_object_type))

        if object_type == 'vmware':

            sla_id = self.object_id(sla, "sla", timeout=timeout)

            all_vms_in_sla = self.get(
                "v1",
                "/vmware/vm?effective_sla_domain_id={}&is_relic=false".format(sla_id),
                timeout=timeout)

            vm_name_id = {}
            for vm in all_vms_in_sla["data"]:
                vm_name_id[vm["name"]] = vm["id"]

            if bool(vm_name_id) is False:
                raise InvalidParameterException(
                    "The SLA '{}' is currently not protecting any {} objects.".format(
                        sla, object_type))

            return vm_name_id



# Copyright 2018 Rubrik, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

"""
This module contains the Rubrik SDK Connect class.
"""

import base64
import requests
import os
import logging
from random import choice
import time
import socket

from .api import Api
from .cluster import Cluster
from .data_management import Data_Management
from .physical import Physical
from .cloud import Cloud
from .exceptions import InvalidParameterException, RubrikException, APICallException, InvalidTypeException


_CLUSTER = Cluster
_DATA_MANAGEMENT = Data_Management
_PHYSICAL = Physical
_API = Api
_CLOUD = Cloud


class Connect(Cluster, Data_Management, Physical, Cloud):
    """This class acts as the base class for the Rubrik SDK and serves as the main interaction point
    for its end users. It also contains various helper functions used throughout the SDK.

    Arguments:
        _CLUSTER {class} -- This class contains methods related to the management of the Rubrik Cluster itself.
        _DATA_MANAGEMENT {class} - This class contains methods related to backup and restore operations for the various objects managed by the Rubrik Cluster.
        _PHYSICAL {class} - This class contains methods related to the management of the Physical objects in the Rubrik Cluster.
    """

    def __init__(self, node_ip=None, username=None, password=None, api_token=None, enable_logging=False):
        """Constructor for the Connect class which is used to initialize the class variables.

        Keyword Arguments:
            node_ip {str} -- The Hostname or IP Address of a node in the Rubrik cluster you wish to connect to. If a value is not provided we will check for a `rubrik_cdm_node_ip` environment variable. (default: {None})
            username {str} -- The Username you wish to use to connect to the Rubrik cluster. If a value is not provided we will check for a `rubrik_cdm_username` environment variable. (default: {None})
            password {str} -- The Password you wish to use to connect to the Rubrik cluster. If a value is not provided we will check for a `rubrik_cdm_password` environment variable. (default: {None})
            api_token {str} -- The API Token you wish to use to connect to the Rubrik cluster. If populated, the `username` and `password` fields will be ignored. If a value is not provided we will check for a `rubrik_cdm_token` environment variable.  (default: {None})
            enable_logging {bool} -- Flag to determine if logging will be enabled for the SDK. (default: {False})
        """

        if enable_logging:
            logging.getLogger().setLevel(logging.DEBUG)

        if node_ip is None:
            node_ip = os.environ.get('rubrik_cdm_node_ip')
            if node_ip is None:
                raise InvalidParameterException("The Rubrik CDM Node IP has not been provided.")
            else:
                self.node_ip = node_ip
        else:
            self.node_ip = node_ip

        self.log("Node IP: {}".format(self.node_ip))

        # List to store how the credentials have been provided
        credentials_manually_provided = []
        credentials_env_var_provided = []
        # Combined list of manually provided and env var provided
        all_credentials_provided = []

        # Flag used to determine if we have enough information to authenticate against the Rubrik cluster
        credentials_needed_for_authentication = False

        if username:
            credentials_manually_provided.append("username")
        else:
            username = os.environ.get('rubrik_cdm_username')
            if username is not None:
                credentials_env_var_provided.append("username")

        if password:
            credentials_manually_provided.append("password")
        else:
            password = os.environ.get('rubrik_cdm_password')
            if password is not None:
                credentials_env_var_provided.append("password")

        if api_token:
            credentials_manually_provided.append("api_token")
        else:
            api_token = os.environ.get('rubrik_cdm_token')
            if api_token is not None:
                credentials_env_var_provided.append("api_token")

        all_credentials_provided = credentials_manually_provided + credentials_env_var_provided

        if len(credentials_manually_provided) == 3:
            raise InvalidParameterException(
                "You have provided both an API token and a username and password for authentication. You may only use one or the other.")

        if "username" in all_credentials_provided and "password" not in all_credentials_provided:
            raise InvalidParameterException(
                "When providing the username argument, either manually or through the environment variables, you must also provide a password. Alternatively, starting with CDM 5.0, you may also use API Token instead of username and password.")

        if "password" in all_credentials_provided and "username" not in all_credentials_provided:
            raise InvalidParameterException(
                "When providing the password argument, either manually or through the environment variables, you must also provide a username. Alternatively, starting with CDM 5.0, you may also use API Token instead of username and password.")

        if "username" in credentials_manually_provided and "password" in credentials_manually_provided:

            self.username = username
            self.password = password
            self.api_token = None

            self.log("Username: {}".format(self.username))
            self.log("Password: ******")

            credentials_needed_for_authentication = True

        if "api_token" in credentials_manually_provided:
            self.api_token = api_token

            self.log("API Token: ******")

            credentials_needed_for_authentication = True

        if credentials_needed_for_authentication is False:
            if len(credentials_env_var_provided) == 3:
                raise InvalidParameterException(
                    "You have provided both an API token and a username and password, in your environment variables, for authentication. You may only use one or the other.")

            if "username" in all_credentials_provided and "password" in all_credentials_provided:

                self.username = username
                self.password = password
                self.api_token = None

                self.log("Username: {}".format(self.username))
                self.log("Password: ******")

                credentials_needed_for_authentication = True

            if "api_token" in credentials_env_var_provided and credentials_needed_for_authentication is False:
                self.api_token = api_token
                self.log("API Token: ******")
                credentials_needed_for_authentication = True

        if credentials_needed_for_authentication is False:
            raise InvalidParameterException(
                "You must provide either a username and password or API Token for authentication.")

    @staticmethod
    def log(log_message):
        """Create properly formatted debug log messages.

        Arguments:
            log_message {str} -- The message to pass to the debug log.
        """
        log = logging.getLogger(__name__)
        log.debug(log_message)

    def _authorization_header(self):
        """Internal method used to create the authorization header used in the API calls.

        Returns:
            dict -- The authorization header that utilizes Basic authentication.
        """

        if self.api_token is None:
            self.log("Creating the authorization header using the provided username and password.")

            credentials = '{}:{}'.format(self.username, self.password)

            # Encode the Username:Password as base64
            authorization = base64.b64encode(credentials.encode())
            # Convert to String for API Call
            authorization = authorization.decode()

            authorization_header = {
                'Content-Type': 'application/json',
                'Accept': 'application/json',
                'Authorization': 'Basic ' + authorization,
                'User-Agent': 'Rubrik Python SDK v1.0.13'
            }

        else:

            self.log("Creating the authorization header using the provided API Token.")
            authorization_header = {
                'Content-Type': 'application/json',
                'Accept': 'application/json',
                'Authorization': 'Bearer ' + self.api_token,
                'User-Agent': 'Rubrik Python SDK v1.0.13'
            }

        return authorization_header

    @staticmethod
    def _header():
        """Internal method used to create the a header without authorization used in the API calls.

        Returns:
            dict -- The header that does not include any authorization.
        """

        header = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        return header

    @staticmethod
    def _api_validation(api_version, api_endpoint):
        """Internal method used to validate the API Version and API Endpoint provided by the end user

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
        """

        valid_api_versions = ['v1', 'v2', 'internal']

        # Validate the API Version
        if api_version not in valid_api_versions:
            raise InvalidParameterException(
                "Enter a valid API version {}.".format(valid_api_versions))

        # Validate the API Endpoint Syntax
        if not isinstance(api_endpoint, str):
            raise InvalidTypeException("The API Endpoint must be a string.")
        elif api_endpoint[0] != "/":
            raise InvalidParameterException(
                "The API Endpoint should begin with '/'. (ex: /cluster/me)")
        elif api_endpoint[-1] == "/":
            if api_endpoint[-2] != "=":
                raise InvalidParameterException(
                    "Error: The API Endpoint should not end with '/' unless proceeded by '='. (ex. /cluster/me or /fileset/snapshot/<id>/browse?path=/)")


class Bootstrap(_API):
    """This class contains all functions related to the Bootstrapping of a Rubrik Cluster.

    Arguments:
        _API {class} - This class contains the base API methods that can be called independently or internally in standalone functions.
    """

    def __init__(self, node_ip, enable_logging=False):
        """Constructor for the Bootstrap class which is used to initialize the class variables.
        """
        if enable_logging:
            logging.getLogger().setLevel(logging.DEBUG)

        self.node_ip = node_ip
        self.log("User Provided Node IP: {}".format(self.node_ip))
        node_resolution = False
        self.ipv6_addr = ""

        try:
            # Attempt to resolve and/or obtain scope for supplied address
            ip_info = socket.getaddrinfo(self.node_ip, 443, socket.AF_INET6)
            # Extract address from response
            self.ipv6_addr = ip_info[0][4][0]
            if '::ffff' in self.ipv6_addr:
                self.ipv6_addr = ""
                self.log('Resolved IPv4 address')
                # ip_info = socket.getaddrinfo(self.node_ip, 443, socket.AF_INET)
                self.log("Resolved Node IP: {}".format(self.node_ip))
                node_resolution = True
            else:
                self.log('Resolved IPv6 address')
                # Extract scope from response
                self.ipv6_scope = str(ip_info[0][4][3])
                # Properly format link-local IPv6 address with scope
                self.node_ip = ('[{}%{}]').format(self.ipv6_addr, self.ipv6_scope)
                self.log("Resolved Node IP: {}".format(self.node_ip))
                node_resolution = True
        except socket.gaierror:
            self.log('Could not resolve link-local IPv6 address for cluster.')

        # IPv6 resolution failed, verify IPv4
        if node_resolution == False:
            try:
                ip_info = socket.getaddrinfo(self.node_ip, 443, socket.AF_INET)
                self.log("Resolved Node IP: {}".format(self.node_ip))
                node_resolution = True
            except socket.gaierror:
                self.log('Could not resolve IPv4 address for cluster.')

        if node_resolution == False:
            sys.exit(
                "Error: Could not resolve addrsss for cluster, or invalid IP/address supplied "
            )

    def setup_cluster(self, cluster_name, admin_email, admin_password, management_gateway, management_subnet_mask, node_config=None, enable_encryption=True, dns_search_domains=None, dns_nameservers=None, ntp_servers=None, wait_for_completion=True, timeout=30):  # pylint: ignore
        """Issues a bootstrap request to a specified Rubrik cluster

        Arguments:
            cluster_name {str} -- Unique name to assign to the Rubrik cluster.
            admin_email {str} -- The Rubrik cluster sends messages for the admin account to this email address.
            admin_password {str} --  Password for the admin account.
            management_gateway {str} --  IP address assigned to the management network gateway
            management_subnet_mask {str} -- Subnet mask assigned to the management network.

        Keyword Arguments:
            node_config {dict} -- The Node Name and IP formatted as a dictionary. (default: {None})
            enable_encryption {bool} -- Enable software data encryption at rest. When bootstraping a Cloud Cluster this value needs to be False. (default: {True})
            dns_search_domains {str} -- The search domain that the DNS Service will use to resolve hostnames that are not fully qualified. (default: {None})
            dns_nameservers {list} -- IPv4 addresses of DNS servers. (default: {['8.8.8.8']})
            ntp_servers {list} -- FQDN or IPv4 address of a network time protocol (NTP) server. (default: {['pool.ntp.org']})
            wait_for_completion {bool} -- Flag to determine if the function should wait for the bootstrap process to complete. (default: {True})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {30})

        Returns:
            dict -- The response returned by `POST /internal/cluster/me/bootstrap`.
        """

        if node_config is None or isinstance(node_config, dict) is not True:
            raise InvalidTypeException(
                'You must provide a valid dictionary for "node_config".')

        if dns_search_domains is None:
            dns_search_domains = []
        elif isinstance(dns_search_domains, list) is not True:
            raise InvalidTypeException(
                'You must provide a valid list for "dns_search_domains".')

        if dns_nameservers is None:
            dns_nameservers = ['8.8.8.8']
        elif isinstance(dns_nameservers, list) is not True:
            raise InvalidTypeException('You must provide a valid list for "dns_nameservers".')

        if ntp_servers is None:
            ntp_servers = ['pool.ntp.org']
        elif isinstance(ntp_servers, list) is not True:
            raise InvalidTypeException('You must provide a valid list for "ntp_servers".')

        bootstrap_config = {}
        bootstrap_config["enableSoftwareEncryptionAtRest"] = enable_encryption
        bootstrap_config["name"] = cluster_name
        bootstrap_config["dnsNameservers"] = dns_nameservers
        bootstrap_config["dnsSearchDomains"] = dns_search_domains
        bootstrap_config["ntpServers"] = ntp_servers

        bootstrap_config["adminUserInfo"] = {}
        bootstrap_config["adminUserInfo"]['password'] = admin_password
        bootstrap_config["adminUserInfo"]['emailAddress'] = admin_email
        bootstrap_config["adminUserInfo"]['id'] = "admin"

        bootstrap_config["nodeConfigs"] = {}
        for node_name, node_ip in node_config.items():
            bootstrap_config["nodeConfigs"][node_name] = {}
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig'] = {}
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['netmask'] = management_subnet_mask
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['gateway'] = management_gateway
            bootstrap_config["nodeConfigs"][node_name]['managementIpConfig']['address'] = node_ip

        while True:

            try:
                self.log('bootstrap: Starting the bootstrap process.')
                number_of_attempts = 1
                api_request = self.post(
                    'internal',
                    '/cluster/me/bootstrap',
                    bootstrap_config,
                    timeout,
                    authentication=False)
                break
            except APICallException as bootstrap_error:
                if "Failed to establish a new connection: [Errno 111] Connection refused" in str(
                        bootstrap_error):
                    self.log(
                        'bootstrap: Connection refused. Waiting 30 seconds for the node to initialize before trying again.')
                    number_of_attempts += 1
                    time.sleep(30)
                elif "Cannot bootstrap from an already bootstrapped node" in str(bootstrap_error):
                    return "No change required. The Rubrik cluster is already bootstrapped."
                else:
                    self.log('bootstrap: Connection refused.')
                    raise RubrikException(bootstrap_error)

            if number_of_attempts == 12:
                raise APICallException(
                    "Unable to establish a connection to the Rubrik cluster.")

        request_id = api_request['id']

        if wait_for_completion:
            self.log('bootstrap: Waiting for the bootstrap process to complete.')
            while True:
                status = self.status(request_id)

                if status['status'] == 'IN_PROGRESS':
                    self.log("bootstrap_status: {}\n".format(status))
                    time.sleep(30)
                    continue
                elif status['status'] == 'FAILURE' or status['status'] == "FAILED":
                    raise RubrikException("{}".format(status['message']))
                else:
                    self.log("{}".format(status))
                    return status

        return api_request

    def status(self, request_id="1", timeout=15):
        """Retrieves status of in progress bootstrap requests

        Keyword Arguments:
            request_id {str} -- ID of the bootstrap request(default: {"1"})
            timeout {int} -- The response timeout value, in seconds, of the API call. (default: {15})

        Returns:
            dict -- The response returned by `GET /internal/cluster/me/bootstrap?request_id={request_id}`.
        """

        self.log('status: Getting the status of the Rubrik Cluster bootstrap.')
        bootstrap_status_api_endpoint = '/cluster/me/bootstrap?request_id={}'.format(
            request_id)
        self.log(bootstrap_status_api_endpoint)
        api_request = self.get(
            'internal', bootstrap_status_api_endpoint, timeout=timeout, authentication=False)

        return api_request

    @staticmethod
    def log(log_message):
        """Create properly formatted debug log messages.

        Arguments:
            log_message {str} - - The message to pass to the debug log.
        """
        log = logging.getLogger(__name__)
        log.debug(log_message)

    @staticmethod
    def _header():
        """Internal method used to create the a header without authorization used in the API calls.

        Returns:
            dict - - The header that does not include any authorization.
        """

        header = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        return header

    @staticmethod
    def _api_validation(api_version, api_endpoint):
        """Internal method used to validate the API Version and API Endpoint provided by the end user

        Arguments:
            api_version {str} - - The version of the Rubrik CDM API to call.
            api_endpoint {str} - - The endpoint(ex. cluster / me) of the Rubrik CDM API to call.
        """

        valid_api_versions = ['v1', 'internal']

        # Validate the API Version
        if api_version not in valid_api_versions:
            raise InvalidParameterException(
                "Enter a valid API version {}.".format(valid_api_versions))

        # Validate the API Endpoint Syntax
        if not isinstance(api_endpoint, str):
            raise InvalidTypeException("The API Endpoint must be a string.")
        elif api_endpoint[0] != "/":
            raise InvalidParameterException(
                "The API Endpoint should begin with '/'. (ex: /cluster/me)")
        elif api_endpoint[-1] == "/":
            if api_endpoint[-2] != "=":
                raise InvalidParameterException(
                    "Error: The API Endpoint should not end with '/' unless proceeded by '='. (ex. /cluster/me or /fileset/snapshot/<id>/browse?path=/)")



import pytest
from rubrik_cdm.exceptions import InvalidParameterException, CDMVersionException, InvalidTypeException
from rubrik_cdm import Connect


def test_aws_s3_cloudout_invalid_aws_region(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(
            "awsbucketname",
            "archive_name",
            "not_a_valid_aws_region",
            "aws_region",
            "aws_access_key",
            "aws_secret_key")

    error_message = error.value.args[0]

    assert error_message == "The `aws_region` must be one of the following: ['ap-south-1', 'ap-northeast-2', 'ap-southeast-1', 'ap-southeast-2', 'ap-northeast-1', 'ca-central-1', 'cn-north-1', 'cn-northwest-1', 'eu-central-1', 'eu-west-1', 'eu-west-2', 'eu-west-3', 'sa-east-1', 'us-gov-west-1', 'us-west-1', 'us-east-1', 'us-east-2', 'us-west-2']"


def test_aws_s3_cloudout_invalid_storage_class(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(
            "awsbucketname",
            "archive_name",
            "ap-south-1",
            "aws_access_key",
            "aws_secret_key",
            "kms_master_key_id",
            "rsa_key",
            "not_a_valid_storage_class")

    error_message = error.value.args[0]

    assert error_message == "The `storage_class` must be one of the following: ['standard', 'standard_ia', 'reduced_redundancy', 'onezone_ia']"


@pytest.mark.parametrize('archive_name', ["_", "/", "*", "?", "%", ".", ":", "|", "<", ">"])
def test_aws_s3_cloudout_invalid_aws_bucket_name(rubrik, archive_name):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(archive_name)

    error_message = error.value.args[0]

    assert error_message == r"The `aws_bucket_name` may not contain any of the following characters: _\/*?%.:|<>"


def test_aws_s3_cloudout_missing_aws_region(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout("awsbucketname")

    error_message = error.value.args[0]

    assert error_message == "`aws_region` has not been provided."


def test_aws_s3_cloudout_missing_aws_access_key(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout("awsbucketname", "archive_name", "ap-south-1")

    error_message = error.value.args[0]

    assert error_message == "`aws_access_key` has not been provided."


def test_aws_s3_cloudout_missing_aws_secret_key(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout("awsbucketname", "archive_name", "ap-south-1", "aws_access_key")

    error_message = error.value.args[0]

    assert error_message == "`aws_secret_key` has not been provided."


def test_aws_s3_cloudout_missing_kms_and_rsa(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(
            "awsbucketname",
            "archive_name",
            "ap-south-1",
            "aws_access_key",
            "aws_secret_key")

    error_message = error.value.args[0]

    assert error_message == "You must populated either `kms_master_key_id` or `rsa_key`."


def test_aws_s3_cloudout_both_kms_and_rsa_populated(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(
            "awsbucketname",
            "archive_name",
            "ap-south-1",
            "aws_access_key",
            "aws_secret_key",
            "kms_master_key_id",
            "rsa_key")

    error_message = error.value.args[0]

    assert error_message == "Both `kms_master_key_id` or `rsa_key` have been populated. You may only use one."


def test_aws_s3_cloudout_idempotence(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "archive_name",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "string",
                            "vNetId": "string",
                            "securityGroupId": "string",
                            "resourceGroupId": "string"
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    assert rubrik.aws_s3_cloudout("awsbucketname", "archive_name", "ap-south-1", "aws_access_key", "aws_secret_key", None, "rsa_key") \
        == "No change required. The 'archive_name' archival location is already configured on the Rubrik cluster."


def test_aws_s3_cloudout_archive_name_already_exsits(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "archive_name",
                        "accessKey": "string",
                        "bucket": "string",
                        "defaultRegion": "string",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "string",
                            "vNetId": "string",
                            "securityGroupId": "string",
                            "resourceGroupId": "string"
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudout(
            "awsbucketname",
            "archive_name",
            "ap-south-1",
            "aws_access_key",
            "aws_secret_key",
            None,
            "rsa_key")

    error_message = error.value.args[0]

    assert error_message == "Archival location with name 'archive_name' already exists. Please enter a unique `archive_name`."


def test_aws_s3_cloudout(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "string",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "string",
                            "vNetId": "string",
                            "securityGroupId": "string",
                            "resourceGroupId": "string"
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    def mock_post_internal_archive_object_store():
        return {
            "jobInstanceId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_archive_object_store()

    assert rubrik.aws_s3_cloudout("awsbucketname", "archive_name", "ap-south-1", "aws_access_key", "aws_secret_key", None, "rsa_key") \
        == mock_post_internal_archive_object_store()


def test_update_aws_s3_cloudout_invalid_storage_class(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.update_aws_s3_cloudout("current_archive_name", storage_class="not_a_valid_storage_class")

    error_message = error.value.args[0]

    assert error_message == "The `storage_class` must be one of the following: ['standard', 'standard_ia', 'reduced_redundancy', 'onezone_ia']"


def test_update_aws_s3_cloudout_current_archive_name_not_found(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "string",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "string",
                            "vNetId": "string",
                            "securityGroupId": "string",
                            "resourceGroupId": "string"
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.update_aws_s3_cloudout("current_archive_name")

    error_message = error.value.args[0]

    assert error_message == "No S3 archival location with name 'current_archive_name' exists."


def test_update_aws_s3_cloudout(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "current_archive_name",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "string",
                            "vNetId": "string",
                            "securityGroupId": "string",
                            "resourceGroupId": "string"
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    def mock_patch_internal_archive_object_store():
        return {
            "jobInstanceId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    mock_patch = mocker.patch('rubrik_cdm.Connect.patch', autospec=True, spec_set=True)
    mock_patch.return_value = mock_patch_internal_archive_object_store()

    assert rubrik.update_aws_s3_cloudout("current_archive_name", "new_archive_name", "aws_access_key", "aws_secret_key", "standard") \
        == mock_patch_internal_archive_object_store()


def test_aws_s3_cloudon_idempotence(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "archive_name",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "subnet_id",
                            "vNetId": "vpc_id",
                            "securityGroupId": "security_group_id",
                        },
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    assert rubrik.aws_s3_cloudon("archive_name", "vpc_id", "subnet_id", "security_group_id") \
        == "No change required. The 'archive_name' archival location is already configured for CloudOn."


def test_aws_s3_cloudon_archive_name_not_found(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "string",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.aws_s3_cloudon("archive_name", "vpc_id", "subnet_id", "security_group_id")

    error_message = error.value.args[0]

    assert error_message == "The Rubrik cluster does not have an archive location named 'archive_name'."


def test_aws_s3_cloudon(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "S3",
                        "name": "archive_name",
                        "accessKey": "aws_access_key",
                        "bucket": "awsbucketname",
                        "defaultRegion": "ap-south-1",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "storageClass": "STANDARD",
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    def mock_patch_internal_archive_object_store():
        return {
            "jobInstanceId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    mock_patch = mocker.patch('rubrik_cdm.Connect.patch', autospec=True, spec_set=True)
    mock_patch.return_value = mock_patch_internal_archive_object_store()

    assert rubrik.aws_s3_cloudon("archive_name", "vpc_id", "subnet_id", "security_group_id") \
        == mock_patch_internal_archive_object_store()


@pytest.mark.parametrize('container', ["_", "/", "*", "?", "%", ".", ":", "|", "<", ">"])
def test_azure_cloudout_invalid_container(rubrik, container):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.azure_cloudout(container, "azure_access_key", "storage_account_name", "rsa_key")

    error_message = error.value.args[0]

    assert error_message == r"The `container` may not contain any of the following characters: _\/*?%.:|<>"


def test_azure_cloudout_invalid_instance_type(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.azure_cloudout("container", "azure_access_key", "storage_account_name",
                              "rsa_key", instance_type="not_a_valid_instance_type")

    error_message = error.value.args[0]

    assert error_message == "The `instance_type` argument must be one of the following: ['default', 'china', 'germany', 'government']"


def test_azure_cloudout_idempotence(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "Azure",
                        "name": "archive_name",
                        "accessKey": "storage_account_name",
                        "bucket": "container",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    assert rubrik.azure_cloudout("container", "azure_access_key", "storage_account_name", "rsa_key", "archive_name") \
        == "No change required. The 'archive_name' archival location is already configured on the Rubrik cluster."


def test_azure_cloudout_invalid_archive_name(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "Azure",
                        "name": "archive_name",
                        "accessKey": "string",
                        "bucket": "string",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.azure_cloudout("container", "azure_access_key", "storage_account_name", "rsa_key", "archive_name")

    error_message = error.value.args[0]

    assert error_message == "Archival location with name 'archive_name' already exists. Please enter a unique `name`."


def test_azure_cloudout(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "Azure",
                        "name": "string",
                        "accessKey": "storage_account_name",
                        "bucket": "container",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },

                }
            ],
            "total": 1
        }

    def mock_patch_internal_archive_object_store():
        return {
            "jobInstanceId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_patch_internal_archive_object_store()

    assert rubrik.azure_cloudout("container", "azure_access_key", "storage_account_name",
                                 "rsa_key", "archive_name") == mock_patch_internal_archive_object_store()


def test_azure_cloudon_invalid_region(rubrik):

    with pytest.raises(InvalidParameterException) as error:
        rubrik.azure_cloudon(
            "archive_name",
            "container",
            "storage_account_name",
            "application_id",
            "application_key",
            "tenant_id",
            "not_a_valid_region",
            "virtual_network_id",
            "subnet_name",
            "security_group_id")

    error_message = error.value.args[0]

    assert error_message == "The `region` must be one of the following: ['westus', 'westus2', 'centralus', 'eastus', 'eastus2', 'northcentralus', 'southcentralus', 'westcentralus', 'canadacentral', 'canadaeast', 'brazilsouth', 'northeurope', 'westeurope', 'uksouth', 'ukwest', 'eastasia', 'southeastasia', 'japaneast', 'japanwest', 'australiaeast', 'australiasoutheast', 'centralindia', 'southindia', 'westindia', 'koreacentral', 'koreasouth']"


def test_azure_cloudon_idempotence(rubrik, mocker):

    def mock_get_internal_archive_object_store():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "polarisManagedId": "string",
                    "definition": {
                        "objectStoreType": "Azure",
                        "name": "archive_name",
                        "accessKey": "storage_account_name",
                        "bucket": "container",
                        "isComputeEnabled": True,
                        "isConsolidationEnabled": True,
                        "defaultComputeNetworkConfig": {
                            "subnetId": "subnet_name",
                            "vNetId": "/subscriptions/89b90dec-a6e1-4q9e-bc12-23138bb3cee4/resourceGroups/PythonSDK/providers/Microsoft.Network/virtualNetworks/pythonsdk",
                            "securityGroupId": "security_group_id",
                            "resourceGroupId": "string"
                        },
                        "azureComputeSummary": {
                            "tenantId": "tenant_id",
                            "subscriptionId": "89b90dec-a6e1-4q9e-bc12-23138bb3cee4",
                            "clientId": "application_id",
                            "region": "westus",
                            "generalPurposeStorageAccountName": "storage_account_name",
                            "containerName": "container",
                            "environment": "AZURE"
                        },
                        "encryptionType": "RSA_KEY_ENCRYPTION"
                    },


                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_archive_object_store()

    assert rubrik.azure_cloudon(
        "archive_name",
        "container",
        "storage_account_name",
        "application_id",
        "application_key",
        "tenant_id",
        "westus",
        "/subscriptions/89b90dec-a6e1-4q9e-bc12-23138bb3cee4/resourceGroups/PythonSDK/providers/Microsoft.Network/virtualNetworks/pythonsdk",
        "subnet_name",
        "security_group_id") == "No change required. The 'archive_name' archival location is already configured for CloudOn."


def test_update_aws_native_account_minimum_installed_cdm_version_not_met(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.0.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(CDMVersionException) as error:
        rubrik.update_aws_native_account("aws_account_name", {})


def test_update_aws_native_account_invalid_config(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidTypeException) as error:
        rubrik.update_aws_native_account("aws_account_name", "not_a_dictionary")

    error_message = error.value.args[0]

    assert error_message == "The 'config' argument must be a dictionary."


def test_update_aws_native_account(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    def mock_get_internal_aws_account_name():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "name": "aws_account_name",
                    "primaryClusterId": "string",
                    "status": "Connected"
                }
            ],
            "total": 1
        }

    def mock_patch_internal_aws_account_id():
        return {
            "name": "string",
            "accessKey": "string",
            "regions": [
                "string"
            ],
            "regionalBoltNetworkConfigs": [
                {
                    "region": "string",
                    "vNetId": "string",
                    "subnetId": "string",
                    "securityGroupId": "string"
                }
            ],
            "disasterRecoveryArchivalLocationId": "string",
            "id": "string",
            "configuredSlaDomainId": "string",
            "configuredSlaDomainName": "string",
            "primaryClusterId": "string"
        }

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_aws_account_name()

    mock_patch = mocker.patch('rubrik_cdm.Connect.patch', autospec=True, spec_set=True)
    mock_patch.return_value = mock_patch_internal_aws_account_id()

    assert rubrik.update_aws_native_account("aws_account_name", {}) == \
        mock_patch_internal_aws_account_id()


def test_add_aws_native_account_minimum_installed_cdm_version_not_met(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.0.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(CDMVersionException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["not_a_valid_region"],
            {})


def test_add_aws_native_account_missing_aws_region(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account("aws_account_name")

    error_message = error.value.args[0]

    assert error_message == "`aws_region` has not been provided."


def test_add_aws_native_account_missing_aws_access_key(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account("aws_account_name", aws_regions=["ap-south-1"])

    error_message = error.value.args[0]

    assert error_message == "`aws_access_key` has not been provided."


def test_add_aws_native_account_missing_aws_secret_key(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account("aws_account_name", "aws_access_key", aws_regions=["ap-south-1"])

    error_message = error.value.args[0]

    assert error_message == "`aws_secret_key` has not been provided."


def test_add_aws_native_account_invalid_aws_regionss(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["not_a_valid_region"],
            {})

    error_message = error.value.args[0]

    assert error_message == "The list `aws_regions` may only contain the following values: ['ap-south-1', 'ap-northeast-3', 'ap-northeast-2', 'ap-southeast-1', 'ap-southeast-2', 'ap-northeast-1', 'ca-central-1', 'cn-north-1', 'cn-northwest-1', 'eu-central-1', 'eu-west-1', 'eu-west-2', 'eu-west-3', 'us-west-1', 'us-east-1', 'us-east-2', 'us-west-2']"


def test_add_aws_native_account_invalid_regional_bolt_network_configs_list(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidTypeException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["ap-south-1"],
            "not_a_valid_regional_bolt_config")

    error_message = error.value.args[0]

    assert error_message == "`regional_bolt_network_configs` must be a list if defined."


def test_add_aws_native_account_invalid_regional_bolt_network_configs_list_dict(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidTypeException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["ap-south-1"],
            ["not_a_valid_regional_bolt_config"])

    error_message = error.value.args[0]

    assert error_message == "The `regional_bolt_network_configs` list can only contain dicts."


def test_add_aws_native_account_invalid_regional_bolt_network_configs_list_dict_value(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["ap-south-1"],
            [{"not_a_valid_key": "string"}])

    error_message = error.value.args[0]

    assert error_message == "Each `regional_bolt_network_config` dict must contain the following keys: 'region', 'vNetId', 'subnetId', 'securityGroupId'."


def test_add_aws_native_account_invalid_aws_account_name(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    def mock_get_internal_aws_account_name():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "name": "aws_account_name",
                    "primaryClusterId": "string",
                    "status": "Connected"
                }
            ],
            "total": 1
        }

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_aws_account_name()

    with pytest.raises(InvalidParameterException) as error:
        rubrik.add_aws_native_account(
            "aws_account_name",
            "aws_access_key",
            "aws_secret_key",
            ["ap-south-1"],
            [{"region": "string", "vNetId": "string", "subnetId": "string", "securityGroupId": "string"}])

    error_message = error.value.args[0]

    assert error_message == "Cloud native source with name 'aws_account_name' already exists. Please enter a unique `aws_account_name`."


def test_add_aws_native_account_idempotence(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    def mock_get_internal_aws_account_name():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "name": "string",
                    "primaryClusterId": "string",
                    "status": "Connected"
                }
            ],
            "total": 1
        }

    def mock_get_internal_aws_account_id():
        return {
            "name": "string",
            "accessKey": "aws_access_key",
            "regions": [
                "string"
            ],
            "regionalBoltNetworkConfigs": [
                {
                    "region": "string",
                    "vNetId": "string",
                    "subnetId": "string",
                    "securityGroupId": "string"
                }
            ],
            "disasterRecoveryArchivalLocationId": "string",
            "id": "string",
            "configuredSlaDomainId": "string",
            "configuredSlaDomainName": "string",
            "primaryClusterId": "string"
        }

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_get_internal_aws_account_name(), mock_get_internal_aws_account_id()]

    assert rubrik.add_aws_native_account("aws_account_name",
                                         "aws_access_key",
                                         "aws_secret_key",
                                         ["ap-south-1"],
                                         [{"region": "string",
                                           "vNetId": "string",
                                           "subnetId": "string",
                                           "securityGroupId": "string"}]) == "No change required. Cloud native source with access key 'aws_access_key' is already configured on the Rubrik cluster."


def test_add_aws_native_account(rubrik, mocker):

    def mock_self_cluster_version():
        return "4.2.1-1280"

    def mock_get_internal_aws_account_name():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "name": "string",
                    "primaryClusterId": "string",
                    "status": "Connected"
                }
            ],
            "total": 1
        }

    def mock_get_internal_aws_account_id():
        return {
            "name": "string",
            "accessKey": "string",
            "regions": [
                "string"
            ],
            "regionalBoltNetworkConfigs": [
                {
                    "region": "string",
                    "vNetId": "string",
                    "subnetId": "string",
                    "securityGroupId": "string"
                }
            ],
            "disasterRecoveryArchivalLocationId": "string",
            "id": "string",
            "configuredSlaDomainId": "string",
            "configuredSlaDomainName": "string",
            "primaryClusterId": "string"
        }

    def mock_post_internal_aws_account():
        return {
            "id": "string",
            "status": "string",
            "progress": 0,
            "startTime": "2019-05-03T16:15:38.827Z",
            "endTime": "2019-05-03T16:15:38.827Z",
            "nodeId": "string",
            "error": {
                "message": "string"
            },
            "links": [
                {
                    "href": "string",
                    "rel": "string"
                }
            ]
        }
    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_get_internal_aws_account_name(), mock_get_internal_aws_account_id()]

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_aws_account()

    assert rubrik.add_aws_native_account("aws_account_name",
                                         "aws_access_key",
                                         "aws_secret_key",
                                         ["ap-south-1"],
                                         [{"region": "string",
                                           "vNetId": "string",
                                           "subnetId": "string",
                                           "securityGroupId": "string"}]) == mock_post_internal_aws_account()



import pytest
from rubrik_cdm.exceptions import InvalidParameterException, CDMVersionException, InvalidTypeException
from rubrik_cdm import Connect


def test_cluster_version(rubrik, mocker):

    def mock_get_v1_cluster_me_version():
        return {'version': '5.0.1-1280'}

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_v1_cluster_me_version()

    assert rubrik.cluster_version() == "5.0.1-1280"


def test_minimum_installed_cdm_version_met(rubrik, mocker):

    def mock_self_cluster_version():
        return "5.0.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    assert rubrik.minimum_installed_cdm_version("5.0") is True


def test_minimum_installed_cdm_version_not_met(rubrik, mocker):

    def mock_self_cluster_version():
        return "5.0.1-1280"

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = mock_self_cluster_version()

    assert rubrik.minimum_installed_cdm_version("5.2") is False


def test_cluster_node_ip(rubrik, mocker):

    def mock_internal_cluster_me_node():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "192.168.1.1",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "string",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "192.168.1.2",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "string",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "192.168.1.3",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                }

            ],
            "total": 0
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_internal_cluster_me_node()

    assert rubrik.cluster_node_ip() == ["192.168.1.1", "192.168.1.2", "192.168.1.3"]


def test_cluster_node_name(rubrik, mocker):

    def mock_internal_cluster_me_node():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "RVM000A000001",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000002",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000003",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                }

            ],
            "total": 0
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_internal_cluster_me_node()

    assert rubrik.cluster_node_name() == ["RVM000A000001", "RVM000A000002", "RVM000A000003"]


def test_end_user_authorization_invalid_object(rubrik):

    with pytest.raises(InvalidParameterException):
        rubrik.end_user_authorization("object_name", "end_user", "not_a_supported_object_type")


def test_end_user_authorization_invalid_end_user(rubrik, mocker):

    def mock_self_object_id():
        return "VirtualMachine:::e6a7e6f1-6050-1ee33-9ba6-8e284e2801de-vm-38297"

    def mock_internal_user_username():
        return []

    mock_object_id = mocker.patch('rubrik_cdm.Connect.object_id', autospec=True, spec_set=True)
    mock_object_id.return_value = mock_self_object_id

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_internal_user_username()

    with pytest.raises(InvalidParameterException):
        rubrik.end_user_authorization("object_name", "end_user", "vmware")


def test_end_user_authorization_idempotence(rubrik, mocker):

    def mock_self_object_id():
        return "VirtualMachine:::e6a7e6f1-6050-1ee33-9ba6-8e284e2801de-vm-38297"

    def mock_internal_user_username():
        return [
            {
                "id": "User:::119283ae-22ea-13f3-bfe2-9387cdf1d4a",
                "authDomainId": "string",
                "username": "string",
                "firstName": "string",
                "lastName": "string",
                "emailAddress": "string",
                "contactNumber": "string",
                "mfaServerId": "string"
            }
        ]

    def mock_internal_authorization_role_end_user_principals():
        return {
            "hasMore": True,
            "data": [
                {
                    "principal": "string",
                    "privileges": {
                        "destructiveRestore": [
                            "string"
                        ],
                        "restore": [
                            "VirtualMachine:::e6a7e6f1-6050-1ee33-9ba6-8e284e2801de-vm-38297"
                        ],
                        "onDemandSnapshot": [
                            "string"
                        ],
                        "restoreWithoutDownload": [
                            "string"
                        ],
                        "viewEvent": [
                            "string"
                        ],
                        "provisionOnInfra": [
                            "string"
                        ],
                        "viewReport": [
                            "string"
                        ]
                    },
                    "organizationId": "string"
                }
            ],
            "total": 0
        }

    mock_object_id = mocker.patch('rubrik_cdm.Connect.object_id', autospec=True, spec_set=True)
    mock_object_id.return_value = mock_self_object_id()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_internal_user_username(), mock_internal_authorization_role_end_user_principals()]

    assert rubrik.end_user_authorization("object_name", "end_user", "vmware", 1) \
        == 'No change required. The End User "end_user" is already authorized to interact with the "object_name" VM.'


def test_end_user_authorization(rubrik, mocker):

    def mock_self_object_id():
        return "VirtualMachine:::e6a7e6f1-6050-1ee33-9ba6-8e284e2801de-vm-38297"

    def mock_internal_user_username():
        return [
            {
                "id": "User:::119283ae-22ea-13f3-bfe2-9387cdf1d4a",
                "authDomainId": "string",
                "username": "string",
                "firstName": "string",
                "lastName": "string",
                "emailAddress": "string",
                "contactNumber": "string",
                "mfaServerId": "string"
            }
        ]

    def mock_internal_authorization_role_end_user_principals():
        return {
            "hasMore": True,
            "data": [
                {
                    "principal": "string",
                    "privileges": {
                        "destructiveRestore": [
                            "string"
                        ],
                        "restore": [
                            "VirtualMachine:::e6a7e6r3-6050-1ee33-9ba6-8e284e2801de"
                        ],
                        "onDemandSnapshot": [
                            "string"
                        ],
                        "restoreWithoutDownload": [
                            "string"
                        ],
                        "viewEvent": [
                            "string"
                        ],
                        "provisionOnInfra": [
                            "string"
                        ],
                        "viewReport": [
                            "string"
                        ]
                    },
                    "organizationId": "string"
                }
            ],
            "total": 0
        }

    def mock_internal_authorization_role_end_user():
        return {
            "hasMore": False,
            "data": [
                {
                    "principal": "User:::119283ae-22ea-13f3-bfe2-9387cdf1d4a",
                    "privileges": {
                        "destructiveRestore": [],
                        "restore": [
                            "VirtualMachine:::e6a7e6f1-6050-1ee33-9ba6-8e284e2801de-vm-38297-not-present"
                        ],
                        "onDemandSnapshot": [],
                        "restoreWithoutDownload": [],
                        "viewEvent": [],
                        "provisionOnInfra": [],
                        "viewReport": []
                    },
                    "organizationId": "Organization:::05e3ee0b-5ec1-e33b-88a5-d916855aff5f"
                }
            ],
            "total": 1
        }

    mock_object_id = mocker.patch('rubrik_cdm.Connect.object_id', autospec=True, spec_set=True)
    mock_object_id.return_value = mock_self_object_id()

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_internal_user_username(), mock_internal_authorization_role_end_user_principals()]

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_internal_authorization_role_end_user()

    assert rubrik.end_user_authorization("object_name", "end_user", "vmware") \
        == mock_internal_authorization_role_end_user()


def test_add_vcenter_idempotence(rubrik, mocker):

    def mock_v1_vmware_vcenter_primary_cluster_id():
        return {
            "hasMore": True,
            "data": [
                {
                    "caCerts": "string",
                    "configuredSlaDomainId": "string",
                    "id": "string",
                    "name": "string",
                    "configuredSlaDomainName": "string",
                    "primaryClusterId": "string",
                    "hostname": "vCenter-Hostname",
                    "username": "string",
                    "conflictResolutionAuthz": "AllowAutoConflictResolution",
                    "configuredSlaDomainPolarisManagedId": "string"
                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_v1_vmware_vcenter_primary_cluster_id()

    assert rubrik.add_vcenter("vCenter-Hostname", "vcenter_username", "vcenter_password") == \
        "No change required. The vCenter 'vCenter-Hostname' has already been added to the Rubrik cluster."


def test_add_vcenter(rubrik, mocker):

    def mock_v1_vmware_vcenter_primary_cluster_id():
        return {
            "hasMore": True,
            "data": [
                {
                    "caCerts": "string",
                    "configuredSlaDomainId": "string",
                    "id": "string",
                    "name": "string",
                    "configuredSlaDomainName": "string",
                    "primaryClusterId": "string",
                    "hostname": "string",
                    "username": "string",
                    "conflictResolutionAuthz": "AllowAutoConflictResolution",
                    "configuredSlaDomainPolarisManagedId": "string"
                }
            ],
            "total": 1
        }

    def mock_v1_vmware_vcenter():
        return {
            "id": "string",
            "status": "string",
            "progress": 0,
            "startTime": "2019-04-17T02:46:12.097Z",
            "endTime": "2019-04-17T02:46:12.097Z",
            "nodeId": "string",
            "error": {
                "message": "string"
            },
            "links": [
                {
                    "href": "www.example.com",
                    "rel": "string"
                }
            ]
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_v1_vmware_vcenter_primary_cluster_id()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_v1_vmware_vcenter()

    assert rubrik.add_vcenter("vCenter-Hostname", "vcenter_username", "vcenter_password") == \
        (mock_v1_vmware_vcenter(), "www.example.com")


def test_configure_timezone_invalid_timezone(rubrik):

    with pytest.raises(InvalidParameterException):
        rubrik.configure_timezone("not_a_supported_timezone")


def test_configure_timezone_idempotence(rubrik, mocker):

    def mock_get_v1_cluster_me():
        return {
            "id": "string",
            "version": "string",
            "apiVersion": "string",
            "name": "string",
            "timezone": {
                "timezone": "America/Chicago"
            },
            "geolocation": {
                "address": "string"
            },
            "acceptedEulaVersion": "string",
            "latestEulaVersion": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_v1_cluster_me()

    assert rubrik.configure_timezone("America/Chicago") \
        == "No change required. The Rubrik cluster is already configured with 'America/Chicago' as it's timezone."


def test_configure_timezone(rubrik, mocker):

    def mock_get_v1_cluster_me():
        return {
            "id": "string",
            "version": "string",
            "apiVersion": "string",
            "name": "string",
            "timezone": {
                "timezone": "America/Denver"
            },
            "geolocation": {
                "address": "string"
            },
            "acceptedEulaVersion": "string",
            "latestEulaVersion": "string"
        }

    def mock_patch_v1_cluster_me():
        return {
            "id": "string",
            "version": "string",
            "apiVersion": "string",
            "name": "string",
            "timezone": {
                "timezone": "America/Denver"
            },
            "geolocation": {
                "address": "string"
            },
            "acceptedEulaVersion": "string",
            "latestEulaVersion": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_v1_cluster_me()

    mock_patch = mocker.patch('rubrik_cdm.Connect.patch', autospec=True, spec_set=True)
    mock_patch.return_value = mock_patch_v1_cluster_me()

    assert rubrik.configure_timezone("America/Chicago") == mock_patch_v1_cluster_me()


def test_configure_ntp_invalid_type(rubrik):

    with pytest.raises(InvalidTypeException):
        rubrik.configure_ntp("not_a_list")


def test_configure_syslog_invalid_protocol(rubrik):

    with pytest.raises(InvalidParameterException):
        rubrik.configure_syslog("syslog_ip", "not_a_valid_protocol")


def test_configure_syslog_invalid_idempotence(rubrik, mocker):

    def mock_get_internal_syslog():
        return {
            "hasMore": True,
            "data": [
                {
                    "hostname": "syslog_ip",
                    "port": 514,
                    "protocol": "TCP",
                    "id": "string"
                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_syslog()

    assert rubrik.configure_syslog("syslog_ip", "TCP") == \
        "No change required. The Rubrik cluster is already configured to use the syslog server 'syslog_ip' on port '514' using the 'TCP' protocol."


def test_configure_syslog(rubrik, mocker):

    def mock_get_internal_syslog():
        return {
            "hasMore": True,
            "data": [
                {
                    "hostname": "syslog_ip",
                    "port": 514,
                    "protocol": "TCP",
                    "id": "string"
                }
            ],
            "total": 1
        }

    def mock_delete_internal_syslog_id():
        return {'status_code': '204'}

    def mock_post_internal_syslog():
        return {
            "hostname": "syslog_ip_new",
            "port": 514,
            "protocol": "TCP",
            "id": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_syslog()

    mock_delete = mocker.patch('rubrik_cdm.Connect.delete', autospec=True, spec_set=True)
    mock_delete.return_value = mock_delete_internal_syslog_id()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_syslog()

    assert rubrik.configure_syslog("syslog_ip_new", "TCP") == mock_post_internal_syslog()


def test_configure_vlan_invalid_ip(rubrik):

    with pytest.raises(InvalidParameterException):
        rubrik.configure_vlan("vlan", "netmask", "not_valid_a_list_or_dict")


def test_configure_vlan_invalid_number_of_vlans(rubrik, mocker):

    def mock_internal_cluster_me_node():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "RVM000A000001",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000002",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000003",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                }

            ],
            "total": 0
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_internal_cluster_me_node()

    with pytest.raises(InvalidParameterException):
        rubrik.configure_vlan("vlan", "netmask", ["IP_1", "IP_2"])


def test_configure_vlan(rubrik, mocker):

    def mock_internal_cluster_me_node():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "RVM000A000001",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000002",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                },
                {
                    "id": "RVM000A000003",
                    "brikId": "string",
                    "status": "string",
                    "ipAddress": "string",
                    "supportTunnel": {
                        "isTunnelEnabled": True,
                        "port": 0,
                        "enabledTime": "2019-04-16T14:16:15.573Z",
                        "lastActivityTime": "2019-04-16T14:16:15.573Z",
                        "inactivityTimeoutInSeconds": 0
                    }
                }

            ],
            "total": 0
        }

    def mock_internal_cluster_me_vlan():
        return {
            "hasMore": True,
            "data": [
                {
                    "vlan": 0,
                    "netmask": "string",
                    "interfaces": [
                        {
                            "node": "string",
                            "ip": "string"
                        }
                    ]
                }
            ],
            "total": 0
        }

    def mock_post_internal_cluster_me_vlan():
        return {'status_code': '204'}

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_internal_cluster_me_node(), mock_internal_cluster_me_vlan()]

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_cluster_me_vlan()

    assert rubrik.configure_vlan("100", "netmask", ["IP_1", "IP_2", "IP_3"]) == mock_post_internal_cluster_me_vlan()


def test_configure_dns_servers_invalid_server_ip(rubrik):

    with pytest.raises(InvalidTypeException):
        rubrik.configure_dns_servers("not_a_valid_server_ip_type")


def test_configure_dns_servers_idempotence(rubrik, mocker):

    def mock_get_internal_cluster_me_dns_nameserver():
        return [
            "server_1"
        ]

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_cluster_me_dns_nameserver()

    assert rubrik.configure_dns_servers(["server_1"]) == \
        "No change required. The Rubrik cluster is already configured with the provided DNS servers."


def test_configure_dns_servers(rubrik, mocker):

    def mock_get_internal_cluster_me_dns_nameserver():
        return [
            "server_1",
            "server_2"
        ]

    def mock_post_internal_cluster_me_dns_nameserver():
        return {'status_code': '204'}

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_cluster_me_dns_nameserver()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_cluster_me_dns_nameserver()

    assert rubrik.configure_dns_servers(["server_1"]) == mock_post_internal_cluster_me_dns_nameserver()


def test_configure_search_domain_invalid_search_domain(rubrik):

    with pytest.raises(InvalidTypeException):
        rubrik.configure_search_domain("not_a_valid_search_domain_type")


def test_configure_search_domain_idempotence(rubrik, mocker):

    def mock_get_internal_cluster_me_dns_search_domain():
        return [
            "domain.1",
        ]

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_cluster_me_dns_search_domain()

    assert rubrik.configure_search_domain(["domain.1"]) == \
        "No change required. The Rubrik cluster is already configured with the provided DNS Search Domains."


def test_configure_search_domain(rubrik, mocker):

    def mock_get_internal_cluster_me_dns_search_domain():
        return [
            "server_1",
        ]

    def mock_post_internal_cluster_me_dns_search_domain():
        return {'status_code': '204'}

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_cluster_me_dns_search_domain()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_cluster_me_dns_search_domain()

    assert rubrik.configure_search_domain(["domain.1"]) == mock_post_internal_cluster_me_dns_search_domain()


def test_configure_smtp_settings_invalid_encryption(rubrik):

    with pytest.raises(InvalidParameterException):
        rubrik.configure_smtp_settings("hostname", "port", "from_email", "smtp_username",
                                       "smtp_password", "not_a_valid_encryption_value")


def test_configure_smtp_settings_idempotence(rubrik, mocker):

    def mock_get_internal_smtp_instance():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "smtpHostname": "hostname",
                    "smtpPort": 0,
                    "smtpSecurity": "NONE",
                    "smtpUsername": "smtp_username",
                    "fromEmailId": "from_email"
                }
            ],
            "total": 1
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_smtp_instance()

    assert rubrik.configure_smtp_settings("hostname", "0", "from_email", "smtp_username", "smtp_password", "NONE") == \
        "No change required. The Rubrik cluster is already configured with the provided SMTP settings."


def test_configure_smtp_settings_new(rubrik, mocker):

    def mock_get_internal_smtp_instance():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "smtpHostname": "string",
                    "smtpPort": 0,
                    "smtpSecurity": "string",
                    "smtpUsername": "string",
                    "fromEmailId": "string"
                }
            ],
            "total": 0
        }

    def mock_post_internal_smtp_instance():
        return {
            "id": "string",
            "smtpHostname": "string",
            "smtpPort": 0,
            "smtpSecurity": "string",
            "smtpUsername": "string",
            "fromEmailId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_smtp_instance()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_smtp_instance()

    assert rubrik.configure_smtp_settings("hostname", "0", "from_email", "smtp_username", "smtp_password", "NONE") == \
        mock_post_internal_smtp_instance()


def test_configure_smtp_settings_update(rubrik, mocker):

    def mock_get_internal_smtp_instance():
        return {
            "hasMore": True,
            "data": [
                {
                    "id": "string",
                    "smtpHostname": "string",
                    "smtpPort": 0,
                    "smtpSecurity": "string",
                    "smtpUsername": "string",
                    "fromEmailId": "string"
                }
            ],
            "total": 1
        }

    def mock_patch_internal_smtp_instance_id():
        return {
            "id": "string",
            "smtpHostname": "string",
            "smtpPort": 0,
            "smtpSecurity": "string",
            "smtpUsername": "string",
            "fromEmailId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_smtp_instance()

    mock_patch = mocker.patch('rubrik_cdm.Connect.patch', autospec=True, spec_set=True)
    mock_patch.return_value = mock_patch_internal_smtp_instance_id()

    assert rubrik.configure_smtp_settings("hostname", "0", "from_email", "smtp_username", "smtp_password", "NONE") == \
        mock_patch_internal_smtp_instance_id()


def test_refresh_vcenter_no_wait(rubrik, mocker):

    def mock_object_id():
        return "vCenter:::eeeb7c90-a074-1233-e6e6-90386f8c3d70"

    def mock_post_v1_vmware_vcenter_id_refresh():
        return {
            "id": "string",
            "status": "string",
            "progress": 0,
            "startTime": "2019-04-17T21:22:14.214Z",
            "endTime": "2019-04-17T21:22:14.214Z",
            "nodeId": "string",
            "error": {
                "message": "string"
            },
            "links": [
                {
                    "href": "build.rubrik.com",
                    "rel": "string"
                }
            ]
        }

    mock_get_object_id = mocker.patch('rubrik_cdm.Connect.object_id', autospec=True, spec_set=True)
    mock_get_object_id.return_value = mock_object_id()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_v1_vmware_vcenter_id_refresh()

    assert rubrik.refresh_vcenter("vcenter_ip", False) == mock_post_v1_vmware_vcenter_id_refresh()


def test_refresh_vcenter_wait_for_completion(rubrik, mocker):

    def mock_object_id():
        return "vCenter:::eeeb7c90-a074-1233-e6e6-90386f8c3d70"

    def mock_job_status():
        return {
            "id": "REFRESH_METADATA_eeeeb7c81-a074-4525-b5b5-81796f8c3d70_0d3e10b0-6741-4c1a-e208-c661325c1efd:::0",
            "status": "SUCCEEDED",
            "startTime": "2019-04-17T21:31:17.785Z",
            "endTime": "2019-04-17T21:31:39.056Z",
            "nodeId": "cluster:::RVM189S019012",
            "links": [
                {
                    "href": "REFRESH_METADATA_eeeeb7c81-a074-4525-b5b5-81796f8c3d70_0d3e10b0-6741-4c1a-e208-c661325c1efd:::0",
                    "rel": "self"
                }
            ]
        }

    mock_get_object_id = mocker.patch('rubrik_cdm.Connect.object_id', autospec=True, spec_set=True)
    mock_get_object_id.return_value = mock_object_id()

    mock_common_api = mocker.patch('rubrik_cdm.Connect._common_api', autospec=True, spec_set=True)
    mock_common_api.return_value = mock_job_status()

    assert rubrik.refresh_vcenter("vcenter_ip", True) == mock_job_status()


def test_create_user_idempotence(rubrik, mocker):

    def mock_get_internal_user():
        return [
            {
                "id": "string",
                "authDomainId": "string",
                "username": "string",
                "firstName": "string",
                "lastName": "string",
                "emailAddress": "string",
                "contactNumber": "string",
                "mfaServerId": "string"
            }
        ]

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_user()

    assert rubrik.create_user("username", "password") == \
        "No change required. The user 'username' already exists on the Rubrik cluster."


def test_create_user(rubrik, mocker):

    def mock_get_internal_user():
        return []

    def mock_post_internal_user():
        return {
            "id": "string",
            "authDomainId": "string",
            "username": "username",
            "firstName": "string",
            "lastName": "string",
            "emailAddress": "string",
            "contactNumber": "string",
            "createdById": "string",
            "createTime": "string",
            "mfaServerId": "string"
        }

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_user()

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_user()

    assert rubrik.create_user("username", "password") == mock_post_internal_user()


def test_read_only_authorization_minimum_installed_cdm_version(rubrik, mocker):

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = "4.1.2"

    with pytest.raises(CDMVersionException):
        rubrik.read_only_authorization("username")


def test_read_only_authorization_invalid_user(rubrik, mocker):

    def mock_get_internal_user():
        return []

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = "5.0.1"

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.return_value = mock_get_internal_user()

    with pytest.raises(InvalidParameterException):
        rubrik.read_only_authorization("username")


def test_read_only_authorization_idempotence(rubrik, mocker):

    def mock_get_internal_user():
        return [
            {
                "id": "string",
                "authDomainId": "string",
                "username": "string",
                "firstName": "string",
                "lastName": "string",
                "emailAddress": "string",
                "contactNumber": "string",
                "mfaServerId": "string"
            }
        ]

    def mock_get_internal_authorization_role_read_only_admin_principals():
        return {
            "hasMore": True,
            "data": [
                {
                    "principal": "string",
                    "privileges": {
                        "basic": [
                            "Global:::All"
                        ]
                    },
                    "organizationId": "string"
                }
            ],
            "total": 0
        }

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = "5.0.1"

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_get_internal_user(),
                            mock_get_internal_authorization_role_read_only_admin_principals()]

    assert rubrik.read_only_authorization("username") == \
        "No change required. The user 'username' already has read-only permissions."


def test_read_only_authorization(rubrik, mocker):

    def mock_get_internal_user():
        return [
            {
                "id": "string",
                "authDomainId": "string",
                "username": "string",
                "firstName": "string",
                "lastName": "string",
                "emailAddress": "string",
                "contactNumber": "string",
                "mfaServerId": "string"
            }
        ]

    def mock_get_internal_authorization_role_read_only_admin_principals():
        return {
            "hasMore": True,
            "data": [
                {
                    "principal": "string",
                    "privileges": {
                        "basic": [
                            "string"
                        ]
                    },
                    "organizationId": "string"
                }
            ],
            "total": 0
        }

    def mock_post_internal_authorization_role_read_only_admin():
        return {
            "hasMore": True,
            "data": [
                {
                    "principal": "string",
                    "privileges": {
                        "basic": [
                            "string"
                        ]
                    },
                    "organizationId": "string"
                }
            ],
            "total": 1
        }

    mock_cluster_version = mocker.patch('rubrik_cdm.Connect.cluster_version', autospec=True, spec_set=True)
    mock_cluster_version.return_value = "5.0.1"

    mock_get = mocker.patch('rubrik_cdm.Connect.get', autospec=True, spec_set=True)
    mock_get.side_effect = [mock_get_internal_user(),
                            mock_get_internal_authorization_role_read_only_admin_principals()]

    mock_post = mocker.patch('rubrik_cdm.Connect.post', autospec=True, spec_set=True)
    mock_post.return_value = mock_post_internal_authorization_role_read_only_admin()

    assert rubrik.read_only_authorization("username") == mock_post_internal_authorization_role_read_only_admin()



 # Copyright 2018 Rubrik, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

"""
This module contains the Rubrik SDK API class.
"""

import requests
import json
import time
try:
    from urllib import quote  # Python 2.X
except ImportError:
    from urllib.parse import quote  # Python 3+
from random import choice
from .exceptions import APICallException, InvalidParameterException, RubrikException, InvalidTypeException


class Api():
    """This class contains the base API methods that can be called independently or internally in standalone functions."""

    def __init__(self, node_ip):
        super().__init__(node_ip)

    def _common_api(self, call_type, api_version, api_endpoint, config=None, job_status_url=None, timeout=15, authentication=True, params=None):  # pylint: ignore
        """Internal method that consolidates the base API functions.

        Arguments:
            call_type {str} -- The HTTP Method for the type of RESTful API call being made. (choices: {'GET', 'POST', 'PATCH', 'DELETE', and 'JOB_STATUS'.})
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            params {dict} -- An optional dict containing variables in a key:value format to send with `GET` & `DELETE` API calls (default: {None})
            config {dict} -- The specified data to send with 'DELETE', `POST` and `PATCH` API calls. (default: {None})
            job_status_url {str} -- The job status URL provided by a previous API call. (default: {None})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The full API call response for the provided endpoint.
        """

        # Determine if authentication should be sent as part of the API Header
        if authentication:
            header = self._authorization_header()
        elif authentication is False:
            header = self._header()
        else:
            raise InvalidTypeException('"authentication" must be either True or False')

        # Create required header for the special case of a bootstrap including Host attribute
        if call_type != 'JOB_STATUS':
            self._api_validation(api_version, api_endpoint)

            if '/cluster/me/bootstrap' in api_endpoint:
                if self.ipv6_addr != "":
                    header = {
                        'Content-Type': 'application/json',
                        'Accept': 'application/json',
                        'Host': '[' + self.ipv6_addr + ']'
                    }
                    self.log('Created boostrap header: ' + str(header))
                else:
                    header = {
                        'Content-Type': 'application/json',
                        'Accept': 'application/json',
                    }
                    self.log('Created boostrap header: ' + str(header))

        try:
            # Determine which call type is being used and then set the relevant
            # variables for that call type
            if call_type == 'GET':
                request_url = "https://{}/api/{}{}".format(self.node_ip, api_version, api_endpoint)
                if params is not None:
                    request_url = request_url + "?" + '&'.join("{}={}".format(key, val)
                                                               for (key, val) in params.items())
                request_url = quote(request_url, '://?=&[]')
                self.log('GET {}'.format(request_url))
                api_request = requests.get(
                    request_url, verify=False, headers=header, timeout=timeout)
            elif call_type == 'POST':
                config = json.dumps(config)
                request_url = "https://{}/api/{}{}".format(
                    self.node_ip, api_version, api_endpoint)
                self.log('POST {}'.format(request_url))
                self.log('Config: {}'.format(config))
                api_request = requests.post(request_url, verify=False, headers=header, data=config, timeout=timeout)
                self.log('Response: {}'.format(api_request.text))
            elif call_type == 'PATCH':
                config = json.dumps(config)
                request_url = "https://{}/api/{}{}".format(
                    self.node_ip, api_version, api_endpoint)
                self.log('PATCH {}'.format(request_url))
                self.log('Config: {}'.format(config))
                api_request = requests.patch(request_url, verify=False, headers=header, data=config, timeout=timeout)
            elif call_type == 'DELETE':
                config = json.dumps(config)
                request_url = "https://{}/api/{}{}".format(
                    self.node_ip, api_version, api_endpoint)
                if params is not None:
                    request_url = request_url + "?" + '&'.join("{}={}".format(key, val)
                                                               for (key, val) in params.items())
                self.log('DELETE {}'.format(request_url))
                api_request = requests.delete(request_url, verify=False, headers=header, data=config, timeout=timeout)
            elif call_type == 'JOB_STATUS':
                self.log('JOB STATUS for {}'.format(job_status_url))
                api_request = requests.get(job_status_url, verify=False, headers=header, timeout=timeout)
            else:
                raise InvalidParameterException('the _common_api() call_type must be one of the following: {}'.format(
                    ['GET', 'POST', 'PATCH', 'DELETE', 'JOB_STATUS']))

            self.log(str(api_request) + "\n")
            try:
                api_response = api_request.json()
                # Check to see if an error message has been provided by Rubrik
                for key, value in api_response.items():
                    if key == "errorType" or key == 'message':
                        error_message = api_response['message']
                        api_request.raise_for_status()

            except BaseException:
                api_request.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise APICallException("Unable to establish a connection to the Rubrik cluster.")
        except requests.exceptions.ConnectionError:
            raise APICallException("Unable to establish a connection to the Rubrik cluster.")
        except requests.exceptions.ReadTimeout:
            raise APICallException(
                "The Rubrik cluster did not respond to the API request in the allotted amount of time. To fix this issue, increase the timeout value.")
        except requests.exceptions.RequestException as error:
            # If "error_message" has be defined raise that message else raise the request exception error
            try:
                error_message
            except NameError:
                raise APICallException(error)
            else:
                raise APICallException('' + error_message)
        else:
            try:
                return api_request.json()
            except BaseException:
                return {'status_code': api_request.status_code}

    def get(self, api_version, api_endpoint, timeout=15, authentication=True, params=None):
        """Send a GET request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            params {dict} -- An optional dict containing variables in a key:value format to send with `GET` & `DELETE` API calls (default: {None})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api(
            'GET',
            api_version,
            api_endpoint,
            config=None,
            job_status_url=None,
            timeout=timeout,
            authentication=authentication,
            params=params)

    def post(self, api_version, api_endpoint, config, timeout=15, authentication=True):
        """Send a POST request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
            config {dict} -- The specified data to send with the API call.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api(
            'POST',
            api_version,
            api_endpoint,
            config=config,
            job_status_url=None,
            timeout=timeout,
            authentication=authentication)

    def patch(self, api_version, api_endpoint, config, timeout=15, authentication=True):
        """Send a PATCH request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).
            config {dict} -- The specified data to send with the API call.

        Keyword Arguments:
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """

        return self._common_api(
            'PATCH',
            api_version,
            api_endpoint,
            config=config,
            job_status_url=None,
            timeout=timeout,
            authentication=authentication)

    def delete(self, api_version, api_endpoint, timeout=15, authentication=True, config=None, params=None):
        """Send a DELETE request to the provided Rubrik API endpoint.

        Arguments:
            api_version {str} -- The version of the Rubrik CDM API to call. (choices: {v1, v2, internal})
            api_endpoint {str} -- The endpoint of the Rubrik CDM API to call (ex. /cluster/me).

        Keyword Arguments:
            params {dict} -- An optional dict containing variables in a key:value format to send with `GET` & `DELETE` API calls . Mutually exclusive with config (default: {None})
            config {dict} -- The specified data to send with the API call. Mutually exclusive with params (default: {None})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})
            authentication {bool} -- Flag that specifies whether or not to utilize authentication when making the API call. (default: {True})

        Returns:
            dict -- The response body of the API call.
        """
        if config is not None and params is not None:
            raise InvalidParameterException(
                "DELETE cannot have both params and config set in the same call.")

        return self._common_api(
            'DELETE',
            api_version,
            api_endpoint,
            config=config,
            job_status_url=None,
            timeout=timeout,
            authentication=authentication,
            params=params)

    def job_status(self, url, wait_for_completion=True, timeout=15):
        """Certain Rubrik operations (on-demand snapshots, live mounts, etc.) may not complete instantaneously. In those cases we have
        the ability to monitor the status of the job through a job status url provided in the actions API response body. This function will
        perform a GET operation on the provided url and return the jobs status.

        Arguments:
            url {str} -- The job status URL provided by a previous API call.

        Keyword Arguments:
            wait_for_completion {bool} -- Flag that determines if the method should wait for the job to complete before exiting. (default: {True})
            timeout {int} -- The number of seconds to wait to establish a connection the Rubrik cluster before returning a timeout error. (default: {15})


        Returns:
            dict -- The response body of the API call.
        """

        if not isinstance(wait_for_completion, bool):
            raise InvalidTypeException(
                'The job_status() wait_for_completion argument must be True or False.')

        if wait_for_completion:
            self.log('Job Status: Waiting for the job to complete.')
            api_call = self._common_api(
                'JOB_STATUS',
                api_version=None,
                api_endpoint=None,
                config=None,
                job_status_url=url,
                timeout=timeout)

            while True:

                api_call = self._common_api(
                    'JOB_STATUS',
                    api_version=None,
                    api_endpoint=None,
                    config=None,
                    job_status_url=url,
                    timeout=timeout)

                job_status = api_call['status']

                if job_status == "SUCCEEDED":
                    self.log('Job Progress 100%\n')
                    job_status = api_call['status']
                    break
                elif job_status == "QUEUED" or "RUNNING":
                    self.log('Job Progress {}%\n'.format(api_call['progress']))
                    job_status = api_call['status']
                    time.sleep(10)
                    continue
                else:
                    raise RubrikException('{}'.format(str(api_call)))

        else:
            api_call = self._common_api(
                'JOB_STATUS',
                api_version=None,
                api_endpoint=None,
                config=None,
                job_status_url=url,
                timeout=timeout)

        return api_call



# Copyright 2019 Rubrik, Inc.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.


class RubrikException(Exception):
    """Base class for exceptions in this module."""
    pass


class CDMVersionException(RubrikException):
    """Exception used to handle situations when the Rubrik cluster is not running a minimum
    required version of CDM.

    Arguments:
        RubrikException {class} -- Base class for exceptions in this module
    """

    def __init__(self, cdm_version):
        self.cdm_version = cdm_version

    def __str__(self):
        return("The Rubrik cluster must be running CDM version {} or later.".format(self.cdm_version))


class APICallException(RubrikException):
    """Exception related to the underlying API call being made to the Rubrik cluster."""
    pass


class InvalidParameterException(RubrikException):
    """Exception related to the parameters provided in the function. This can be related an issue with the value itself
    or the value provided not being found on the cluster."""
    pass


class InvalidTypeException(RubrikException):
    """Exception related to the wrong Python type being provided in the function parameters."""
    pass
